"""Build- and meta-related mixins that isolate shareable functionality between Hera objects."""

from __future__ import annotations

import functools
import inspect
import sys
from collections import ChainMap
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Type, TypeVar, Union, cast

if sys.version_info >= (3, 10):
    from inspect import get_annotations
    from types import NoneType
    from typing import Concatenate, ParamSpec
else:
    from typing_extensions import Concatenate, ParamSpec

    from hera.shared._inspect import get_annotations

    NoneType = type(None)


from hera.shared import BaseMixin, global_config
from hera.shared._global_config import _DECORATOR_SYNTAX_FLAG, _flag_enabled
from hera.shared._pydantic import BaseModel, get_fields, root_validator
from hera.shared._type_util import construct_io_from_annotation, get_annotated_metadata, unwrap_annotation
from hera.workflows._context import _context
from hera.workflows.exceptions import InvalidTemplateCall
from hera.workflows.io.v1 import (
    Input as InputV1,
    Output as OutputV1,
)
from hera.workflows.models import (
    Artifact as ModelArtifact,
    Parameter as ModelParameter,
    TemplateRef,
)
from hera.workflows.parameter import Parameter
from hera.workflows.protocol import TWorkflow

try:
    from hera.workflows.io.v2 import (  # type: ignore
        Input as InputV2,
        Output as OutputV2,
    )
except ImportError:
    from hera.workflows.io.v1 import (  # type: ignore
        Input as InputV2,
        Output as OutputV2,
    )


if TYPE_CHECKING:
    from hera.workflows._mixins import TemplateMixin
    from hera.workflows.steps import Step
    from hera.workflows.task import Task

_yaml: Optional[ModuleType] = None
try:
    import yaml

    _yaml = yaml
except ImportError:
    _yaml = None


_varname_imported: bool = False
try:
    # user must install `hera[experimental]`
    from varname import ImproperUseError, varname

    _varname_imported = True
except ImportError:
    pass

THookable = TypeVar("THookable", bound="HookMixin")
"""`THookable` is the type associated with mixins that provide the ability to apply hooks from the global config"""

TContext = TypeVar("TContext", bound="ContextMixin")
"""`TContext` is the bounded context controlled by the context mixin that enable context management in workflow/dag"""


class ContextMixin(BaseMixin):
    """`ContextMixin` provides the ability to implement context management.

    The mixin implements the `__enter__` and `__exit__` functionality that enables the core `with` clause. The mixin
    expects that inheritors implement the `_add_sub` functionality, which adds a node defined within the context to the
    main object context such as `Workflow`, `DAG`, or `ContainerSet`.
    """

    def __enter__(self: TContext) -> TContext:
        """Enter the context of the inheritor."""
        _context.enter(self)
        return self

    def __exit__(self, *_) -> None:
        """Leave the context of the inheritor."""
        _context.exit()

    def _add_sub(self, node: Any) -> Any:
        """Adds the supplied node to the context of the inheritor."""
        raise NotImplementedError()


class ExperimentalMixin(BaseMixin):
    _experimental_warning_message: str = (
        "Unable to instantiate {} since it is an experimental feature."
        " Please turn on experimental features by setting "
        '`hera.shared.global_config.experimental_features["{}"] = True`.'
        " Note that experimental features are unstable and subject to breaking changes."
    )

    _flag: str

    @root_validator(allow_reuse=True)
    def _check_enabled(cls, values):
        if not global_config.experimental_features[cls._flag]:
            raise ValueError(cls._experimental_warning_message.format(cls, cls._flag))
        return values


def _set_model_attr(model: BaseModel, attrs: List[str], value: Any):
    # The `attrs` list represents a path to an attribute in `model`, whose attributes
    # are BaseModels themselves. Therefore we use `getattr` to get a reference to the final
    # BaseModel set to `curr`, then call `setattr` on that BaseModel, using the last attribute
    # name in attrs, and the value passed in.
    curr: BaseModel = model
    for attr in attrs[:-1]:
        curr = getattr(curr, attr)

    setattr(curr, attrs[-1], value)


def _get_model_attr(model: BaseModel, attrs: List[str]) -> Any:
    # This is almost the same as _set_model_attr.
    # The `attrs` list represents a path to an attribute in `model`, whose attributes
    # are BaseModels themselves. Therefore we use `getattr` to get a reference to the final
    # BaseModel set to `curr`, then `getattr` on that BaseModel, using the last attribute
    # name in attrs.
    curr: BaseModel = model
    for attr in attrs[:-1]:
        curr = getattr(curr, attr)

    return getattr(curr, attrs[-1])


class ModelMapperMixin(BaseMixin):
    """`ModelMapperMixin` allows Hera classes to be mapped to auto-generated Argo classes."""

    class ModelMapper:
        def __init__(self, model_path: str, hera_builder: Optional[Callable] = None):
            self.model_path = []
            self.builder = hera_builder

            if not model_path:
                # Allows overriding parent attribute annotations to remove the mapping
                return

            self.model_path = model_path.split(".")
            curr_class: Type[BaseModel] = self._get_model_class()
            for key in self.model_path:
                fields = get_fields(curr_class)
                if key not in fields:
                    raise ValueError(f"Model key '{key}' does not exist in class {curr_class}")
                curr_class = fields[key].outer_type_  # type: ignore

        @classmethod
        def _get_model_class(cls) -> Type[BaseModel]:
            raise NotImplementedError

        @classmethod
        def build_model(
            cls, hera_class: Type[ModelMapperMixin], hera_obj: ModelMapperMixin, model: TWorkflow
        ) -> TWorkflow:
            assert isinstance(hera_obj, ModelMapperMixin)

            for attr, annotation in hera_class._get_all_annotations().items():
                if mappers := get_annotated_metadata(annotation, ModelMapperMixin.ModelMapper):
                    if len(mappers) != 1:
                        raise ValueError("Expected only one ModelMapper")

                    # Value comes from builder function if it exists on hera_obj, otherwise directly from the attr
                    value = (
                        getattr(hera_obj, mappers[0].builder.__name__)()
                        if mappers[0].builder is not None
                        else getattr(hera_obj, attr)
                    )
                    if value is not None:
                        _set_model_attr(model, mappers[0].model_path, value)

            return model

    @classmethod
    def _get_all_annotations(cls) -> ChainMap:
        """Gets all annotations of this class and any parent classes."""
        return ChainMap(*(get_annotations(c) for c in cls.__mro__))

    @classmethod
    def _from_model(cls, model: BaseModel) -> ModelMapperMixin:
        """Parse from given model to cls's type."""
        hera_obj = cls()

        for attr, annotation in cls._get_all_annotations().items():
            if mappers := get_annotated_metadata(annotation, ModelMapperMixin.ModelMapper):
                if len(mappers) != 1:
                    raise ValueError("Expected only one model mapper")
                if mappers[0].model_path:
                    value = _get_model_attr(model, mappers[0].model_path)
                    if value is not None:
                        setattr(hera_obj, attr, value)

        return hera_obj

    @classmethod
    def _from_dict(cls, model_dict: Dict, model: Type[BaseModel]) -> ModelMapperMixin:
        """Parse from given model_dict, using the given model type to call its parse_obj."""
        model_workflow = model.parse_obj(model_dict)
        return cls._from_model(model_workflow)

    @classmethod
    def from_dict(cls, model_dict: Dict) -> ModelMapperMixin:
        """Parse from given model_dict."""
        raise NotImplementedError

    @classmethod
    def _from_yaml(cls, yaml_str: str, model: Type[BaseModel]) -> ModelMapperMixin:
        """Parse from given yaml string, using the given model type to call its parse_obj."""
        if not _yaml:
            raise ImportError("PyYAML is not installed")
        return cls._from_dict(_yaml.safe_load(yaml_str), model)

    @classmethod
    def from_yaml(cls, yaml_str: str) -> ModelMapperMixin:
        """Parse from given yaml_str."""
        raise NotImplementedError

    @classmethod
    def _from_file(cls, yaml_file: Union[Path, str], model: Type[BaseModel]) -> ModelMapperMixin:
        yaml_file = Path(yaml_file)
        return cls._from_yaml(yaml_file.read_text(encoding="utf-8"), model)

    @classmethod
    def from_file(cls, yaml_file: Union[Path, str]) -> ModelMapperMixin:
        """Parse from given yaml_file."""
        raise NotImplementedError


class HookMixin(BaseMixin):
    """`HookMixin` provides the ability to dispatch hooks set on the global config to any inheritors."""

    def _dispatch_hooks(self: THookable) -> THookable:
        """Dispatches the global hooks on the current object."""
        output = self
        for hook in global_config._get_pre_build_hooks(output):
            output = hook(output)
            if output is None:
                raise RuntimeError(
                    f"Pre-build hook {hook.__name__} returned None."
                    "Please ensure you are returning the output value from the hook."
                )
        return output


def _get_pydantic_input_type(source: Callable) -> Union[None, Type[InputV1], Type[InputV2]]:
    """Returns a Pydantic Input type for the source, if it is using Pydantic IO."""
    function_parameters = inspect.signature(source).parameters
    if len(function_parameters) != 1:
        return None
    parameter = next(iter(function_parameters.values()))
    parameter_type = unwrap_annotation(parameter.annotation)
    if not isinstance(parameter_type, type) or not issubclass(parameter_type, (InputV1, InputV2)):
        return None
    return parameter_type


def _get_unset_source_parameters_as_items(source: Callable) -> List[Parameter]:
    """Get a list of `Parameter` with their values set to corresponding templated item strings.

    We infer that each non-keyword, positional, argument of the given source function is a
    parameter that stems from a fanout. Therefore, each parameter value takes the form of
    `{{item}}` when there's a single argument or `{{item.<argument name>}}` when there are
    other arguments.

    Returns:
    -------
    List[Parameter]
        A list of identified parameters (possibly empty).
    """
    non_default_parameters: List[Parameter] = []
    if pydantic_input := _get_pydantic_input_type(source):
        for parameter in pydantic_input._get_parameters():
            if parameter.default is None:
                non_default_parameters.append(parameter)
    else:
        for p in inspect.signature(source).parameters.values():
            if p.default == inspect.Parameter.empty and p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                # only add positional or keyword arguments that are not set to a default value
                # as the default value ones are captured by the automatically generated `Parameter` fields for positional
                # kwargs. Otherwise, we assume that the user sets the value of the parameter via the `with_param` field
                io = construct_io_from_annotation(p.name, p.annotation)
                if isinstance(io, Parameter) and io.default is None and not io.output:
                    non_default_parameters.append(io)

    if len(non_default_parameters) == 1:
        non_default_parameters[0].value = "{{item}}"
    else:
        for param in non_default_parameters:
            param.value = "{{item." + str(param.name) + "}}"
    return non_default_parameters


def _get_parameters_used_in_with_items(with_items: Any) -> Optional[List[Parameter]]:
    """Get the optional list of Parameters used in with_items, with their values set to corresponding templated item strings.

    The assembled list of `Parameter` contains all the unique parameters identified from the `with_items` list. For
    example, if the `with_items` list contains 3 serializable elements such as
    `[{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}]`, then only 2 `Parameter`s are returned. Namely, only
    `Parameter(name='a')` and `Parameter(name='b')` is returned, with values `{{item.a}}` and `{{item.b}}`,
    respectively. This helps with the parallel/serial processing of the supplied items. If the list is of
    plain values and not key-values, then `None` is returned as there are no parameter substitutions.
    """
    if not with_items:
        return None

    if isinstance(with_items[0], dict):
        if not all(isinstance(item, dict) for item in with_items):
            raise ValueError(
                "List must contain all dictionaries or no dictionaries. "
                "Alternatively, serialize the dictionaries with `hera.shared.serialization.serialize()`."
            )
        # Check all dictionaries have the same keys, values will be serialised so we don't consider them
        if not all(set(item.keys()) == set(with_items[0].keys()) for item in with_items[1:]):
            raise ValueError(
                "All dictionaries in with_items must have the same keys. "
                "Alternatively, serialize the dictionaries with `hera.shared.serialization.serialize()`."
            )

        # Create a Parameter for each key in the dictionaries
        # Special case for dictionary len 1 as the value is just the item itself
        if len(with_items[0]) == 1:
            param_name = list(with_items[0].keys())[0]
            return [Parameter(name=param_name, value="{{item}}")]

        # Just use first dictionary as we checked type and key equality above
        return [Parameter(name=key, value=f"{{{{item.{key}}}}}") for key in with_items[0].keys()]

    else:
        # Ensure no dictionaries in the list
        if any(isinstance(item, dict) for item in with_items):
            raise ValueError(
                "List must contain all dictionaries or no dictionaries. "
                "Alternatively, serialize the dictionaries with `hera.shared.serialization.serialize()`."
            )
    return None


class CallableTemplateMixin(BaseMixin):
    """`CallableTemplateMixin` provides the ability to 'call' the template like a regular Python function.

    The callable template implements the `__call__` method for the inheritor. The `__call__` method supports invoking
    the template as a regular Python function. The call must be executed within an active context, which is a
    `Workflow`, `DAG` or `Steps` context since the call optionally returns a `Step` or a `Task` depending on the active
    context (`None` for `Workflow`, `Step` for `Steps` and `Task` for `DAG`). Note that `Steps` also supports calling
    templates in a parallel steps context via using `Steps(...).parallel()`. When the call is executed and the template
    does not exist on the active context, i.e. the workflow, it is automatically added for the user. Note that invoking
    the same template multiple times does *not* result in the creation/addition of the same template to the active
    context/workflow. Rather, a union is performed, so space is saved for users on the templates field and templates are
    not duplicated.
    """

    def __call__(self, *args, **kwargs) -> Union[None, Step, Task]:
        if "name" not in kwargs:
            kwargs["name"] = self.name  # type: ignore

        arguments = self._get_arguments(**kwargs)
        parameter_argument_names = self._get_parameter_names(arguments)
        artifact_argument_names = self._get_artifact_names(arguments)

        # when the `source` is set via an `@script` decorator, it does not come in with the `kwargs` so we need to
        # set it here in order for the following logic to capture it
        if "source" not in kwargs and hasattr(self, "source"):
            kwargs["source"] = self.source  # type: ignore

        if "source" in kwargs and "with_param" in kwargs:
            arguments += self._get_templated_source_args(
                parameter_argument_names,
                artifact_argument_names,
                kwargs["source"],
            )
        elif "source" in kwargs and "with_items" in kwargs:
            arguments += self._get_templated_arguments_from_items(
                parameter_argument_names,
                kwargs["with_items"],
                kwargs["source"],
            )

        # it is possible for the user to pass `arguments` via `kwargs` along with `with_param`. The `with_param`
        # additional parameters are inferred and have to be added to the `kwargs['arguments']`, otherwise
        # the step/task will miss adding them when building the final arguments
        kwargs["arguments"] = arguments

        from hera.workflows.dag import DAG
        from hera.workflows.script import Script
        from hera.workflows.steps import Parallel, Step, Steps
        from hera.workflows.task import Task
        from hera.workflows.workflow import Workflow

        if _context.pieces:
            if isinstance(_context.pieces[-1], Workflow):
                # Notes on callable templates under a Workflow:
                # * If the user calls a script directly under a Workflow (outside of a Steps/DAG) then we add the script
                #   template to the workflow and return None.
                # * Containers, ContainerSets and Data objects (i.e. subclasses of CallableTemplateMixin) are already
                #   added when initialized under the Workflow context so a callable doesn't make sense in that context,
                #   so we raise an InvalidTemplateCall exception.
                # * We do not currently validate the added templates to stop a user adding the same template multiple
                #   times, which can happen if "calling" the same script multiple times to add it to the workflow,
                #   or initializing a second `Container` exactly like the first.
                if isinstance(self, Script):
                    _context.add_sub_node(self)
                    return None

                raise InvalidTemplateCall(
                    f"Callable Template '{self.name}' is not callable under a Workflow"  # type: ignore
                )
            if isinstance(_context.pieces[-1], (Steps, Parallel)):
                return Step(template=self, **kwargs)

            if isinstance(_context.pieces[-1], DAG):
                return Task(template=self, **kwargs)

        raise InvalidTemplateCall(
            f"Callable Template '{self.name}' is not under a Workflow, Steps, Parallel, or DAG context"  # type: ignore
        )

    def _get_arguments(self, **kwargs) -> List:
        """Returns a list of arguments from the kwargs given to the template call."""
        kwargs_arguments = kwargs.get("arguments", [])
        kwargs_arguments = kwargs_arguments if isinstance(kwargs_arguments, List) else [kwargs_arguments]  # type: ignore
        return kwargs_arguments

    def _get_parameter_names(self, arguments: List) -> Set[str]:
        """Returns the union of parameter names from the given arguments' parameter objects and dictionary keys."""
        parameters = [arg for arg in arguments if isinstance(arg, (ModelParameter, Parameter))]
        keys = [arg for arg in arguments if isinstance(arg, dict)]
        return {p.name for p in parameters}.union(
            set(functools.reduce(lambda x, y: cast(List[str], x) + list(y.keys()), keys, []))
        )

    def _get_artifact_names(self, arguments: List) -> Set[str]:
        """Returns the set of artifact names that are currently set on the mixin inheritor."""
        from hera.workflows.artifact import Artifact

        artifacts = [arg for arg in arguments if isinstance(arg, (ModelArtifact, Artifact))]
        return {a if isinstance(a, str) else a.name for a in artifacts if a.name}

    def _get_templated_source_args(
        self,
        parameter_argument_names: Set[str],
        artifact_argument_names: Set[str],
        source: Callable,
    ) -> List[Parameter]:
        """Get list of inferred arguments to set from the given source.

        Argo uses the `inputs` field to indicate the expected parameters of a specific template whereas the
        `arguments` are used to indicate exactly what _values_ are assigned to the set inputs. Here,
        we infer the arguments when `with_param` is used. If a source is passed along with `with_param`, we
        infer the arguments to set from the given source. It is assumed that `with_param` will return the
        expected result for Argo to fan out the task on.

        Parameters
        ----------
        parameter_argument_names: Set[str]
            Set of parameter names already constructed from arguments passed in.
        artifact_argument_names: Set[str]
            Set of artifact names already constructed from arguments passed in.
        source: Callable
            The source function to infer the arguments from.

        Returns:
        -------
        List[Parameter]
            The list of inferred arguments to set.
        """
        new_arguments = []
        source_params_as_items = _get_unset_source_parameters_as_items(source)
        for p in source_params_as_items:
            if p.name not in parameter_argument_names and p.name not in artifact_argument_names:
                new_arguments.append(p)
        return new_arguments

    def _get_templated_arguments_from_items(
        self,
        parameter_argument_names: Set[str],
        items: Any,
        source: Callable,
    ) -> List[Parameter]:
        """Infer arguments from the given items.

        The main difference between `with_items` and `with_param` is that param is a serialized version of
        `with_items`. Hence, `with_items` comes in the form of a list of objects, whereas `with_param` comes
        in as a single serialized object. Here, we can infer the parameters to create based on the content
        of `with_items`.

        Parameters
        ----------
        parameter_names: Set[str]
            Set of already constructed parameter names.
        items: Any
            The items to infer the arguments from.
        source: Callable
            The source function to confirm parameter names from.

        Returns:
        -------
        List[Parameter]
            The list of inferred arguments to set.
        """
        item_params = _get_parameters_used_in_with_items(items)
        if item_params is None:
            return []

        source_param_names = {name for name in inspect.signature(source).parameters.keys()}
        new_params = []
        for p in item_params:
            if p.name not in source_param_names:
                raise ValueError(
                    f"Parameter '{p.name}' not found in source function '{source.__name__}' "
                    "(you may need to serialize the dictionaries with `hera.shared.serialization.serialize()`)"
                )

            if p.name not in parameter_argument_names:
                new_params.append(p)
        return new_params


FuncIns = ParamSpec("FuncIns")  # For input types of given func to script decorator
FuncR = TypeVar("FuncR")  # For return type of given func to script decorator

PydanticKwargs = ParamSpec("PydanticKwargs")  # For input attributes of a Pydantic class
PydanticCls = TypeVar("PydanticCls")
StepIns = ParamSpec("StepIns")  # For input attributes of Step class
TaskIns = ParamSpec("TaskIns")  # For input attributes of Task class


def _add_type_hints(
    _pydantic_obj: Callable[PydanticKwargs, PydanticCls],
) -> Callable[
    ...,
    Callable[
        Concatenate[
            object,  # this is the `self` parameter
            PydanticKwargs,  # this adds type hints to the underlying *library* function kwargs
        ],
        Callable[  # we will return a function that is a decorator
            [Callable[FuncIns, FuncR]],  # taking underlying *user* function
            Callable[FuncIns, FuncR],  # and returning it
        ],
    ],
]:
    """Adds type hints to the decorated function."""
    return lambda func: func


class HeraBuildObj:
    def __init__(self, subnode_type: str, output_class: Type[Union[OutputV1, OutputV2]]) -> None:
        self.subnode_type = subnode_type
        self.output_class = output_class


class TemplateDecoratorFuncsMixin(ContextMixin):
    from hera.workflows.container import Container
    from hera.workflows.dag import DAG
    from hera.workflows.script import Script
    from hera.workflows.steps import Steps

    @staticmethod
    def _check_if_enabled(decorator_name: str):
        if not _flag_enabled(_DECORATOR_SYNTAX_FLAG):
            raise ValueError(
                str(
                    "Unable to use {} decorator since it is an experimental feature."
                    " Please turn on experimental features by setting "
                    '`hera.shared.global_config.experimental_features["{}"] = True`.'
                    " Note that experimental features are unstable and subject to breaking changes."
                ).format(decorator_name, _DECORATOR_SYNTAX_FLAG)
            )
        if not _varname_imported:
            raise ImportError(
                "`varname` is not installed. Install `hera[experimental]` to bring in the extra dependency"
            )

    def _create_subnode(
        self,
        subnode_name: str,
        func: Callable,
        template: TemplateMixin,
        *args,
        **kwargs,
    ) -> Union[Step, Task]:
        from hera.workflows.cluster_workflow_template import ClusterWorkflowTemplate
        from hera.workflows.dag import DAG
        from hera.workflows.steps import Parallel, Step, Steps
        from hera.workflows.task import Task
        from hera.workflows.workflow_template import WorkflowTemplate

        subnode_args = None
        if len(args) == 1 and isinstance(args[0], (InputV1, InputV2)):
            subnode_args = args[0]._get_as_arguments()

        if input_in_kwargs := [k for k, v in kwargs.items() if isinstance(v, (InputV1, InputV2))]:
            raise SyntaxError(
                f"Found Input argument(s) in kwargs: {input_in_kwargs}. Input must be passed as a positional-only argument."
            )

        signature = inspect.signature(func)
        output_class = signature.return_annotation

        subnode: Union[Step, Task]

        assert _context.pieces

        template_ref = None
        _context.declaring = False
        if _context.pieces[0] != self and isinstance(self, WorkflowTemplate):
            # Using None for cluster_scope means it won't appear in the YAML spec (saving some bytes),
            # as cluster_scope=False is the default value
            template_ref = TemplateRef(
                name=self.name,
                template=template.name,
                cluster_scope=True if isinstance(self, ClusterWorkflowTemplate) else None,
            )
            # Set template to None as it cannot be set alongside template_ref
            template = None  # type: ignore

        if isinstance(_context.pieces[-1], (Steps, Parallel)):
            subnode = Step(
                name=subnode_name,
                template=template,
                template_ref=template_ref,
                arguments=subnode_args,
                **kwargs,
            )
        elif isinstance(_context.pieces[-1], DAG):
            subnode = Task(
                name=subnode_name,
                template=template,
                template_ref=template_ref,
                arguments=subnode_args,
                depends=" && ".join(sorted(_context.pieces[-1]._current_task_depends)) or None,
                **kwargs,
            )
            _context.pieces[-1]._current_task_depends.clear()

        subnode._build_obj = HeraBuildObj(subnode._subtype, output_class)
        _context.declaring = True
        return subnode

    @_add_type_hints(Script)  # type: ignore
    def script(self, **script_kwargs) -> Callable:
        """A decorator that wraps a function into a Script object.

        Using this decorator users can define a function that will be executed as a script in a container. Once the
        `Script` is returned users can use it as they generally use a `Script` e.g. as a callable inside a DAG or Steps.
        Note that invoking the function will result in the template associated with the script to be added to the
        workflow context, so users do not have to worry about that.

        Parameters
        ----------
        **script_kwargs
            Keyword arguments to be passed to the Script object.

        Returns:
        -------
        Callable
            Function wrapper that holds a `Script` and allows the function to be called to create a Step or Task if
            in a Steps or DAG context.
        """
        self._check_if_enabled("script")

        from hera.workflows.script import RunnerScriptConstructor, Script

        def script_decorator(func: Callable[FuncIns, FuncR]) -> Callable:
            """The internal decorator function.

            Parameters
            ----------
            func: Callable
                Function to wrap.

            Returns:
            -------
            Callable
                Callable that calls the `Script` object `__call__` method when in a Steps or DAG context,
                otherwise calls function itself.
            """
            # instance methods are wrapped in `staticmethod`. Hera can capture that type and extract the underlying
            # function for remote submission since it does not depend on any class or instance attributes, so it is
            # submittable
            if isinstance(func, staticmethod):
                source: Callable = func.__func__
            else:
                source = func

            # take the client-provided `name` if it is submitted, pop the name for otherwise there will be two
            # kwargs called `name`
            # otherwise populate the `name` from the function name
            name = script_kwargs.pop("name", source.__name__.replace("_", "-"))

            if "constructor" not in script_kwargs and "constructor" not in global_config._get_class_defaults(Script):
                script_kwargs["constructor"] = RunnerScriptConstructor()

            signature = inspect.signature(func)
            func_inputs = signature.parameters
            if len(func_inputs) > 1:
                raise SyntaxError("script decorator must be used with a single `Input` arg, or no args.")

            if len(func_inputs) == 1:
                func_input = list(func_inputs.values())[0].annotation
                if not issubclass(func_input, (InputV1, InputV2)):
                    raise SyntaxError("script decorator must be used with a single `Input` arg, or no args.")

            # Open (Workflow) context to add `Script` object automatically
            with self:
                script_template = Script(name=name, source=source, **script_kwargs)

            if not isinstance(script_template.constructor, RunnerScriptConstructor):
                raise ValueError(f"Script '{name}' must use RunnerScriptConstructor")

            @functools.wraps(func)
            def script_call_wrapper(*args, **kwargs) -> Union[FuncR, Step, Task, None]:
                """Invokes a CallableTemplateMixin's `__call__` method using the given SubNode (Step or Task) args/kwargs."""
                if _context.declaring:
                    subnode_name = kwargs.pop("name", None)
                    if not subnode_name:
                        try:
                            # ignore decorator function assignment
                            subnode_name = varname()
                        except ImproperUseError:
                            # Template is being used without variable assignment (so use function name or provided name)
                            subnode_name = script_template.name  # type: ignore

                        assert isinstance(subnode_name, str)
                        subnode_name = subnode_name.replace("_", "-")

                    return self._create_subnode(subnode_name, func, script_template, *args, **kwargs)

                if _context.pieces:
                    return script_template.__call__(*args, **kwargs)

                return func(*args)

            # Set the wrapped function to the original function so that we can use it later
            script_call_wrapper.wrapped_function = func  # type: ignore
            # Set the template name to the inferred name
            script_call_wrapper.template_name = name  # type: ignore

            return script_call_wrapper

        return script_decorator

    @_add_type_hints(Container)  # type: ignore
    def container(self, **container_kwargs) -> Callable:
        self._check_if_enabled("container")
        from hera.workflows.container import Container

        def container_decorator(func: Callable[FuncIns, FuncR]) -> Callable:
            name = container_kwargs.pop("name", func.__name__.replace("_", "-"))

            signature = inspect.signature(func)
            func_inputs = signature.parameters
            inputs = []
            if len(func_inputs) >= 1:
                input_arg = list(func_inputs.values())[0].annotation
                if issubclass(input_arg, (InputV1, InputV2)):
                    inputs = input_arg._get_inputs(add_missing_path=True)

            func_return = signature.return_annotation
            outputs = []
            if func_return and issubclass(func_return, (OutputV1, OutputV2)):
                outputs = func_return._get_outputs(add_missing_path=True)

            # Open context to add `Container` object automatically
            with self:
                container_template = Container(
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    **container_kwargs,
                )

            @functools.wraps(func)
            def container_call_wrapper(*args, **kwargs) -> Union[FuncR, Step, Task, None]:
                """Invokes a CallableTemplateMixin's `__call__` method using the given SubNode (Step or Task) args/kwargs."""
                if _context.declaring:
                    subnode_name = kwargs.pop("name", None)
                    if not subnode_name:
                        try:
                            # ignore decorator function assignment
                            subnode_name = varname()
                        except ImproperUseError:
                            # Template is being used without variable assignment (so use function name or provided name)
                            subnode_name = container_template.name  # type: ignore

                        assert isinstance(subnode_name, str)
                        subnode_name = subnode_name.replace("_", "-")

                    return self._create_subnode(subnode_name, func, container_template, *args, **kwargs)

                if _context.pieces:
                    return container_template.__call__(*args, **kwargs)
                return func(*args)

            # Set the template name to the inferred name
            container_call_wrapper.template_name = name  # type: ignore

            return container_call_wrapper

        return container_decorator

    def _construct_invocator_decorator(
        self,
        invocator_type: Type[Union[Steps, DAG]],
        **template_kwargs,
    ):
        def decorator(func: Callable[FuncIns, FuncR]) -> Callable:
            """The decorating function that runs at "definition" time.

            This function will do 3 things:
            1. Inspect the function signature to get inputs/outputs
            2. Create the DAG/Steps template and add it to the workflow
            3. Run `func` in "declaring" mode to gather the tasks/steps and inputs/outputs between them, explained below.

            To run the function in declaring mode, we first need to create an input object which will
            carry templated string arguments in its attributes, rather than the correct types (like ints,
            other BaseModels, etc). This will let the templated strings propagate to the task arguments. We
            must skip Pydantic's validation of the input object, which is why InputMixin overrides __new__,
            to `construct` an instance; __init__ then returns early to skip validation.

            Then, passing in the object, we call the underlying function. This is where other templates
            are called, such as scripts, containers or other DAGs, and results in `Task` (or `Step`) objects being
            created. These tasks may have attribute access on them when passing values between tasks, as the
            code author sees Inputs/Outputs, while we are seeing Tasks in declaring mode. Therefore,
            TemplateInvocatorSubNodeMixin overrides __getattribute__, which, when in declaring mode, will
            retrieve a templated string for the given attribute, e.g. `my_task.an_output_param` will get
            the string "{{tasks.my_task.outputs.parameters.an_output_param}}". This also works for artifacts
            and the special `result` output.

            Finally, for the return from the function, a user may specify a subclass of Output. OutputMixin
            therefore also overrides __new__ and __init__ to allow an instance to be constructed without
            validation. We take this output object and convert it to a list of Artifacts/Parameters, setting
            the outputs of the template.
            """
            name = template_kwargs.pop("name", func.__name__.replace("_", "-"))

            signature = inspect.signature(func)
            func_inputs = signature.parameters
            inputs = []
            if len(func_inputs) > 1:
                raise SyntaxError(
                    f"{invocator_type.__name__.lower()} decorator must be used with a single `Input` arg, or no args."
                )

            if len(func_inputs) == 1:
                input_arg = list(func_inputs.values())[0].annotation
                if not issubclass(input_arg, (InputV1, InputV2)):
                    raise SyntaxError(
                        f"{invocator_type.__name__.lower()} decorator must be used with a single `Input` arg, or no args."
                    )
                inputs = input_arg._get_inputs()

            func_return_annotation = signature.return_annotation
            outputs = []
            if func_return_annotation and issubclass(func_return_annotation, (OutputV1, OutputV2)):
                outputs = func_return_annotation._get_outputs()
            elif func_return_annotation is not inspect.Signature.empty and func_return_annotation is not None:
                raise SyntaxError(
                    f"{invocator_type.__name__.lower()} decorator must be used with a single "
                    "`Output` return annotation, or a `None`/empty return annotation."
                )

            # Add dag/steps to workflow
            with self:
                template = invocator_type(
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    **template_kwargs,
                )

            @functools.wraps(func)
            def call_wrapper(*args, **kwargs):
                if _context.declaring:
                    subnode_name = kwargs.pop("name", None)
                    if not subnode_name:
                        try:
                            # ignore decorator function assignment
                            subnode_name = varname()
                        except ImproperUseError:
                            # Template is being used without variable assignment (so use function name or provided name)
                            subnode_name = name  # type: ignore

                        assert isinstance(subnode_name, str)
                        subnode_name = subnode_name.replace("_", "-")

                    return self._create_subnode(subnode_name, func, template, *args, **kwargs)

                return func(*args)

            call_wrapper.template_name = name  # type: ignore

            # Open workflow context to cross-check task template names
            with self, template:
                input_objs = []
                if len(func_inputs) == 1:
                    input_arg = list(func_inputs.values())[0].annotation
                    if issubclass(input_arg, (InputV1, InputV2)):
                        input_objs.append(input_arg._get_as_templated_arguments())

                # "run" the dag/steps function to collect the tasks/steps
                _context.declaring = True
                func_return = func(*input_objs)
                _context.declaring = False

                if func_return is not None:
                    if func_return_annotation is inspect.Signature.empty or func_return_annotation is None:
                        raise SyntaxError(
                            f"Function returned {func_return.__class__}, expected None "
                            "(the function may be missing a return annotation)."
                        )

                    from hera.workflows.steps import Step
                    from hera.workflows.task import Task

                    if isinstance(func_return, (Step, Task)):
                        # User tried to return an `Output` from another step/task directly
                        raise SyntaxError("Function return must be a new Output object.")

                    if issubclass(func_return_annotation, (OutputV1, OutputV2)):
                        if type(func_return) is not func_return_annotation:
                            raise SyntaxError(
                                "Function return does not match annotation, "
                                f"expected: {func_return_annotation}; got: {func_return.__class__}."
                            )
                        assert isinstance(func_return, func_return_annotation)  # for type-checking

                        if func_return.result or func_return.exit_code:
                            raise SyntaxError(
                                "Cannot set `result` or `exit_code` on Output when used in a dag/steps function."
                            )

                        template.outputs = func_return._get_as_invocator_output()

            return call_wrapper

        return decorator

    @_add_type_hints(DAG)  # type: ignore
    def dag(self, **dag_kwargs) -> Callable:
        self._check_if_enabled("dag")

        from hera.workflows.dag import DAG

        return self._construct_invocator_decorator(DAG, **dag_kwargs)

    @_add_type_hints(Steps)  # type: ignore
    def steps(self, **steps_kwargs) -> Callable:
        self._check_if_enabled("steps")

        from hera.workflows.steps import Steps

        return self._construct_invocator_decorator(Steps, **steps_kwargs)



"""The script module provides the Script class.

See https://argoproj.github.io/argo-workflows/workflow-concepts/#script
for more on scripts.
"""

import ast
import copy
import inspect
import sys
import textwrap
from abc import abstractmethod
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

if sys.version_info >= (3, 10):
    from types import NoneType
else:
    NoneType = type(None)


from typing_extensions import ParamSpec, get_args, get_origin

from hera.expr import g
from hera.shared import BaseMixin, global_config
from hera.shared._global_config import (
    _SCRIPT_PYDANTIC_IO_FLAG,
    _flag_enabled,
)
from hera.shared._pydantic import _PYDANTIC_VERSION, root_validator, validator
from hera.shared._type_util import (
    construct_io_from_annotation,
    get_workflow_annotation,
    is_subscripted,
    origin_type_issupertype,
)
from hera.shared.serialization import serialize
from hera.workflows._context import _context
from hera.workflows._meta_mixins import CallableTemplateMixin
from hera.workflows._mixins import (
    ArgumentsT,
    ContainerMixin,
    EnvIOMixin,
    OneOrMany,
    ResourceMixin,
    TemplateMixin,
    VolumeMountMixin,
)
from hera.workflows.artifact import (
    Artifact,
)
from hera.workflows.io.v1 import (
    Input as InputV1,
    Output as OutputV1,
)

try:
    from hera.workflows.io.v2 import (  # type: ignore
        Input as InputV2,
        Output as OutputV2,
    )
except ImportError:
    from hera.workflows.io.v1 import (  # type: ignore
        Input as InputV2,
        Output as OutputV2,
    )
from hera.workflows.models import (
    ContinueOn,
    EnvVar,
    Inputs as ModelInputs,
    Lifecycle,
    LifecycleHook,
    Outputs as ModelOutputs,
    ScriptTemplate as _ModelScriptTemplate,
    SecurityContext,
    Sequence as _ModelSequence,
    Template as _ModelTemplate,
    TemplateRef,
    ValueFrom,
)
from hera.workflows.parameter import Parameter
from hera.workflows.protocol import Templatable
from hera.workflows.steps import Step
from hera.workflows.task import Task
from hera.workflows.volume import _BaseVolume


class ScriptConstructor(BaseMixin):
    """A ScriptConstructor is responsible for generating the source code for a Script given a python callable.

    This allows users to customize the behaviour of the template that hera generates when a python callable is
    passed to the Script class.

    In order to use your custom ScriptConstructor implementation, you can set it as the Script.constructor field.
    """

    @abstractmethod
    def generate_source(self, instance: "Script") -> str:
        """A function that can inspect the Script instance and generate the source field."""
        raise NotImplementedError

    def transform_values(self, cls: Type["Script"], values: Any) -> Any:
        """A function that will be invoked by the root validator of the Script class."""
        return values

    def transform_script_template_post_build(
        self, instance: "Script", script: _ModelScriptTemplate
    ) -> _ModelScriptTemplate:
        """A hook to transform the generated script template."""
        return script

    def transform_template_post_build(self, instance: "Script", template: _ModelTemplate) -> _ModelTemplate:
        """A hook to transform the generated template."""
        return template


class Script(
    EnvIOMixin,
    CallableTemplateMixin,
    ContainerMixin,
    TemplateMixin,
    ResourceMixin,
    VolumeMountMixin,
):
    """A Script acts as a wrapper around a container.

    In Hera this defaults to a "python:3.9" image specified by global_config.image, which runs a python source
    specified by `Script.source`.
    """

    container_name: Optional[str] = None
    args: Optional[List[str]] = None
    command: Optional[List[str]] = None
    lifecycle: Optional[Lifecycle] = None
    security_context: Optional[SecurityContext] = None
    source: Optional[Union[Callable, str]] = None
    working_dir: Optional[str] = None
    add_cwd_to_sys_path: Optional[bool] = None
    constructor: Optional[Union[str, ScriptConstructor]] = None

    @validator("constructor", always=True)
    @classmethod
    def _set_constructor(cls, v):
        if v is None:
            # TODO: In the future we can insert
            # detection code here to determine
            # the best constructor to use.
            v = InlineScriptConstructor()
        if isinstance(v, ScriptConstructor):
            return v
        assert isinstance(v, str)
        if v.lower() == "inline":
            return InlineScriptConstructor()
        elif v.lower() == "runner":
            return RunnerScriptConstructor()
        raise ValueError(f"Unknown constructor {v}")

    @validator("command", always=True)
    @classmethod
    def _set_command(cls, v):
        return v or global_config.script_command

    @validator("add_cwd_to_sys_path", always=True)
    @classmethod
    def _set_add_cwd_to_sys_path(cls, v):
        if v is None:
            return True

    @root_validator
    @classmethod
    def _constructor_validate(cls, values):
        constructor = values.get("constructor")
        assert isinstance(constructor, ScriptConstructor)
        return constructor.transform_values(cls, values)

    def _build_template(self) -> _ModelTemplate:
        assert isinstance(self.constructor, ScriptConstructor)

        return self.constructor.transform_template_post_build(
            self,
            _ModelTemplate(
                active_deadline_seconds=self.active_deadline_seconds,
                affinity=self.affinity,
                archive_location=self.archive_location,
                automount_service_account_token=self.automount_service_account_token,
                daemon=self.daemon,
                executor=self.executor,
                fail_fast=self.fail_fast,
                host_aliases=self.host_aliases,
                init_containers=self._build_init_containers(),
                inputs=self._build_inputs(),
                memoize=self.memoize,
                metadata=self._build_metadata(),
                metrics=self._build_metrics(),
                name=self.name,
                node_selector=self.node_selector,
                outputs=self._build_outputs(),
                parallelism=self.parallelism,
                plugin=self.plugin,
                pod_spec_patch=self.pod_spec_patch,
                priority=self.priority,
                priority_class_name=self.priority_class_name,
                retry_strategy=self.retry_strategy,
                scheduler_name=self.scheduler_name,
                script=self._build_script(),
                security_context=self.pod_security_context,
                service_account_name=self.service_account_name,
                sidecars=self._build_sidecars(),
                synchronization=self.synchronization,
                timeout=self.timeout,
                tolerations=self.tolerations,
                volumes=self._build_volumes(),
            ),
        )

    def _build_script(self) -> _ModelScriptTemplate:
        assert isinstance(self.constructor, ScriptConstructor)
        if _output_annotations_used(cast(Callable, self.source)) and isinstance(
            self.constructor, RunnerScriptConstructor
        ):
            if not self.constructor.outputs_directory:
                self.constructor.outputs_directory = self.constructor.DEFAULT_HERA_OUTPUTS_DIRECTORY
            if self.constructor.volume_for_outputs is not None:
                if self.constructor.volume_for_outputs.mount_path is None:
                    self.constructor.volume_for_outputs.mount_path = self.constructor.outputs_directory
                self._create_hera_outputs_volume(self.constructor.volume_for_outputs)
        assert self.image
        return self.constructor.transform_script_template_post_build(
            self,
            _ModelScriptTemplate(
                args=self.args,
                command=self.command,
                env=self._build_env(),
                env_from=self._build_env_from(),
                image=self.image,
                # `image_pull_policy` in script wants a string not an `ImagePullPolicy` object
                image_pull_policy=self._build_image_pull_policy(),
                lifecycle=self.lifecycle,
                liveness_probe=self.liveness_probe,
                name=self.container_name,
                ports=self.ports,
                readiness_probe=self.readiness_probe,
                resize_policy=self.resize_policy,
                resources=self._build_resources(),
                restart_policy=self.restart_policy,
                security_context=self.security_context,
                source=self.constructor.generate_source(self),
                startup_probe=self.startup_probe,
                stdin=self.stdin,
                stdin_once=self.stdin_once,
                termination_message_path=self.termination_message_path,
                termination_message_policy=self.termination_message_policy,
                tty=self.tty,
                volume_devices=self.volume_devices,
                volume_mounts=self._build_volume_mounts(),
                working_dir=self.working_dir,
            ),
        )

    def _build_inputs(self) -> Optional[ModelInputs]:
        inputs = super()._build_inputs()
        func_parameters: List[Parameter] = []
        func_artifacts: List[Artifact] = []
        if callable(self.source):
            func_parameters, func_artifacts = _get_inputs_from_callable(self.source)

        return cast(Optional[ModelInputs], self._aggregate_callable_io(inputs, func_parameters, func_artifacts, False))

    def _build_outputs(self) -> Optional[ModelOutputs]:
        outputs = super()._build_outputs()

        if not callable(self.source):
            return outputs

        outputs_directory = None
        if isinstance(self.constructor, RunnerScriptConstructor):
            outputs_directory = self.constructor.outputs_directory or self.constructor.DEFAULT_HERA_OUTPUTS_DIRECTORY

        out_parameters, out_artifacts = _get_outputs_from_return_annotation(self.source, outputs_directory)
        func_parameters, func_artifacts = _get_outputs_from_parameter_annotations(self.source, outputs_directory)
        func_parameters.extend(out_parameters)
        func_artifacts.extend(out_artifacts)

        return cast(
            Optional[ModelOutputs], self._aggregate_callable_io(outputs, func_parameters, func_artifacts, True)
        )

    def _aggregate_callable_io(
        self,
        current_io: Optional[Union[ModelInputs, ModelOutputs]],
        func_parameters: List[Parameter],
        func_artifacts: List[Artifact],
        output: bool,
    ) -> Union[ModelOutputs, ModelInputs, None]:
        """Aggregate the Inputs/Outputs with parameters and artifacts extracted from a callable."""
        if not func_parameters and not func_artifacts:
            return current_io
        if current_io is None:
            if output:
                return ModelOutputs(
                    parameters=[p.as_output() for p in func_parameters] or None,
                    artifacts=[a._build_artifact() for a in func_artifacts] or None,
                )

            return ModelInputs(
                parameters=[p.as_input() for p in func_parameters] or None,
                artifacts=[a._build_artifact() for a in func_artifacts] or None,
            )

        seen_params = {p.name for p in current_io.parameters or []}
        seen_artifacts = {a.name for a in current_io.artifacts or []}

        for param in func_parameters:
            if param.name not in seen_params and param.name not in seen_artifacts:
                if current_io.parameters is None:
                    current_io.parameters = []
                if output:
                    current_io.parameters.append(param.as_output())
                else:
                    current_io.parameters.append(param.as_input())

        for artifact in func_artifacts:
            if artifact.name not in seen_artifacts:
                if current_io.artifacts is None:
                    current_io.artifacts = []
                current_io.artifacts.append(artifact._build_artifact())

        return current_io

    def _create_hera_outputs_volume(self, volume: _BaseVolume) -> None:
        """Add given volume to the script template for the automatic saving of the hera outputs."""
        assert isinstance(self.constructor, RunnerScriptConstructor)

        if self.volumes is None:
            self.volumes = []
        elif isinstance(self.volumes, Sequence):
            self.volumes = list(self.volumes)
        elif not isinstance(self.volumes, list):
            self.volumes = [self.volumes]

        if volume not in self.volumes:
            self.volumes.append(volume)


def _get_outputs_from_return_annotation(
    source: Callable,
    outputs_directory: Optional[str],
) -> Tuple[List[Parameter], List[Artifact]]:
    """Returns a tuple of output Parameters and Artifacts defined in the function signature's return annotation."""
    parameters = []
    artifacts = []

    def append_annotation(annotation: Union[Artifact, Parameter]):
        if isinstance(annotation, Artifact):
            if annotation.path is None and outputs_directory is not None:
                annotation.path = outputs_directory + f"/artifacts/{annotation.name}"
            artifacts.append(annotation)
        elif isinstance(annotation, Parameter):
            if annotation.value_from is None and outputs_directory is not None:
                annotation.value_from = ValueFrom(path=outputs_directory + f"/parameters/{annotation.name}")
            parameters.append(annotation)

    return_annotation = inspect.signature(source).return_annotation
    if param_or_artifact := get_workflow_annotation(return_annotation):
        append_annotation(param_or_artifact)
    elif get_origin(return_annotation) is tuple:
        for annotation in get_args(return_annotation):
            if isinstance(annotation, type) and issubclass(annotation, (OutputV1, OutputV2)):
                raise ValueError("Output cannot be part of a tuple output")

            if param_or_artifact := get_workflow_annotation(annotation):
                append_annotation(param_or_artifact)
    elif isinstance(return_annotation, type) and issubclass(return_annotation, (OutputV1, OutputV2)):
        if not _flag_enabled(_SCRIPT_PYDANTIC_IO_FLAG):
            raise ValueError(
                (
                    "Unable to instantiate {} since it is an experimental feature."
                    " Please turn on experimental features by setting "
                    '`hera.shared.global_config.experimental_features["{}"] = True`.'
                    " Note that experimental features are unstable and subject to breaking changes."
                ).format(return_annotation, _SCRIPT_PYDANTIC_IO_FLAG)
            )

        output_class = return_annotation
        for output in output_class._get_outputs():
            append_annotation(output)

    return parameters, artifacts


def _get_outputs_from_parameter_annotations(
    source: Callable,
    outputs_directory: Optional[str],
) -> Tuple[List[Parameter], List[Artifact]]:
    """Returns a tuple of output Parameters and Artifacts defined in the function signature parameters."""
    parameters: List[Parameter] = []
    artifacts: List[Artifact] = []

    for name, p in inspect.signature(source).parameters.items():
        annotation = construct_io_from_annotation(name, p.annotation)
        if not annotation.output:
            continue

        if isinstance(annotation, Parameter) and annotation.value_from is None and outputs_directory is not None:
            annotation.value_from = ValueFrom(path=outputs_directory + f"/parameters/{annotation.name}")
        elif isinstance(annotation, Artifact) and annotation.path is None and outputs_directory is not None:
            annotation.path = outputs_directory + f"/artifacts/{annotation.name}"

        if isinstance(annotation, Artifact):
            artifacts.append(annotation)
        elif isinstance(annotation, Parameter):
            parameters.append(annotation)

    return parameters, artifacts


def _get_inputs_from_callable(source: Callable) -> Tuple[List[Parameter], List[Artifact]]:
    """Return all inputs from the function.

    This includes all basic Python function parameters, and all parameters with a Parameter or Artifact annotation.
    For the Pydantic IO experimental feature, any input parameter which is a subclass of Input, the fields of the
    class will be used as inputs, rather than the class itself.

    Note, the given Parameter/Artifact names in annotations of different inputs could clash, which will raise a ValueError.
    """
    parameters = []
    artifacts = []

    for func_param in inspect.signature(source).parameters.values():
        if not is_subscripted(func_param.annotation) and issubclass(func_param.annotation, (InputV1, InputV2)):
            if not _flag_enabled(_SCRIPT_PYDANTIC_IO_FLAG):
                raise ValueError(
                    (
                        "Unable to instantiate {} since it is an experimental feature."
                        " Please turn on experimental features by setting "
                        '`hera.shared.global_config.experimental_features["{}"] = True`.'
                        " Note that experimental features are unstable and subject to breaking changes."
                    ).format(func_param.annotation, _SCRIPT_PYDANTIC_IO_FLAG)
                )

            if len(inspect.signature(source).parameters) != 1:
                raise SyntaxError("Only one function parameter can be specified when using an Input.")

            input_class = func_param.annotation
            if (
                func_param.default != inspect.Parameter.empty
                and func_param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            ):
                object_override = func_param.default
                parameters.extend(input_class._get_parameters(object_override=object_override))
            else:
                parameters.extend(input_class._get_parameters())

            artifacts.extend(input_class._get_artifacts(add_missing_path=True))

        else:
            io = construct_io_from_annotation(func_param.name, func_param.annotation)
            if io.output:
                continue

            if isinstance(io, Artifact):
                if io.path is None:
                    io.path = io._get_default_inputs_path()

                artifacts.append(io)
            elif isinstance(io, Parameter):
                if io.default is not None:
                    raise ValueError(
                        "default cannot be set via the Parameter's default, use a Python default value instead."
                    )
                if func_param.default != inspect.Parameter.empty:
                    io.default = serialize(func_param.default)

                if origin_type_issupertype(func_param.annotation, NoneType) and io.default != "null":
                    raise ValueError(f"Optional parameter '{func_param.name}' must have a default value of None.")
                parameters.append(io)

    return parameters, artifacts


def _extract_return_annotation_output(source: Callable) -> List:
    """Extract the output annotations from the return annotation of the function signature."""
    output: List[Union[Tuple[type, Union[Parameter, Artifact]], Type[Union[OutputV1, OutputV2]]]] = []

    return_annotation = inspect.signature(source).return_annotation
    origin_type = get_origin(return_annotation)
    annotation_args = get_args(return_annotation)
    if get_workflow_annotation(return_annotation):
        output.append(annotation_args)
    elif origin_type is tuple:
        workflow_args = [
            get_args(annotated_type) for annotated_type in annotation_args if get_workflow_annotation(annotated_type)
        ]

        # If all tuple elements are annotated as Parameter/Artifact
        if len(workflow_args) == len(annotation_args):
            output.extend(workflow_args)
        # Only some tuple elements are annotated as Parameter/Artifact
        elif workflow_args:
            raise ValueError(
                f"Function '{source.__name__}' output has partially annotated tuple return type. "
                "Tuple elements must be all Annotated as Parameter/Artifact, or contain no Parameter/Artifact annotations for a raw tuple return type."
            )
    elif (
        origin_type is None
        and isinstance(return_annotation, type)
        and issubclass(return_annotation, (OutputV1, OutputV2))
    ):
        output.append(return_annotation)

    return output


def _extract_all_output_annotations(source: Callable) -> List:
    """Extract the output annotations out of the function signature.

    This includes parameters marked as outputs and the return annotation of `source`.
    """
    output = []

    for _, func_param in inspect.signature(source).parameters.items():
        io = construct_io_from_annotation(func_param.name, func_param.annotation)
        if io.output:
            output.append(io)

    output.extend(_extract_return_annotation_output(source))

    return output


def _output_annotations_used(source: Callable) -> bool:
    """Check if any output annotations are used.

    This includes parameters marked as outputs and the return annotation of `source`.
    """
    if not callable(source):
        return False

    return bool(_extract_all_output_annotations(source))


FuncIns = ParamSpec("FuncIns")  # For input types of given func to script decorator
FuncR = TypeVar("FuncR")  # For return type of given func to script decorator
FuncRCov = TypeVar("FuncRCov", covariant=True)

ScriptIns = ParamSpec("ScriptIns")  # For input attributes of Script class


class _ScriptDecoratedFunction(Generic[FuncIns, FuncRCov], Protocol):
    """Type assigned to functions decorated with @script."""

    # Note: For more details about overload-overlap, see https://github.com/python/typeshed/issues/12178

    @overload
    def __call__(  # type: ignore [overload-overlap]
        self, *args: FuncIns.args, **kwargs: FuncIns.kwargs
    ) -> FuncRCov:
        # Note: this overload is for calling the decorated function.
        # No docstring is provided, so VS Code will use the docstring of the decorated function.
        ...

    @overload
    def __call__(  # type: ignore [overload-overlap, misc]
        self,
    ) -> Optional[Union[Step, Task]]:
        """Create a Step or Task or add the script as a template to the workflow, depending on the context.

        * Under a DAG context, creates and returns a Task.
        * Under a Steps or Parallel context, creates and returns a Step.
        * Under a Workflow context, adds the script as a template to the Workflow and returns None.

        Use `assert isinstance(result, Step)` or `assert isinstance(result, Task)` to select
        the correct type if using a type-checker.
        """

    @overload
    def __call__(  # type: ignore [overload-overlap]
        self,
        *,
        name: str = ...,
        continue_on: Optional[ContinueOn] = ...,
        hooks: Optional[Dict[str, LifecycleHook]] = ...,
        on_exit: Optional[Union[str, Templatable]] = ...,
        template: Optional[Union[str, _ModelTemplate, TemplateMixin, CallableTemplateMixin]] = ...,
        template_ref: Optional[TemplateRef] = ...,
        inline: Optional[Union[_ModelTemplate, TemplateMixin]] = ...,
        when: Optional[str] = ...,
        with_sequence: Optional[_ModelSequence] = ...,
        arguments: ArgumentsT = ...,
        with_param: Optional[Any] = ...,
        with_items: Optional[OneOrMany[Any]] = ...,
    ) -> Union[Step, Task]:
        """Create a Step or Task, depending on context.

        * Under a DAG context, creates and returns a Task.
        * Under a Steps or Parallel context, creates and returns a Step.

        Use `assert isinstance(result, Step)` or `assert isinstance(result, Task)` to select
        the correct type if using a type-checker.
        """
        # Note: signature must match the Step constructor, except that while name is required for Step,
        # it is automatically inferred from the name of the decorated function when invoked.

    @overload
    def __call__(  # type: ignore [overload-overlap]
        self,
        *,
        name: str = ...,
        continue_on: Optional[ContinueOn] = ...,
        hooks: Optional[Dict[str, LifecycleHook]] = ...,
        on_exit: Optional[Union[str, Templatable]] = ...,
        template: Optional[Union[str, _ModelTemplate, TemplateMixin, CallableTemplateMixin]] = ...,
        template_ref: Optional[TemplateRef] = ...,
        inline: Optional[Union[_ModelTemplate, TemplateMixin]] = ...,
        when: Optional[str] = ...,
        with_sequence: Optional[_ModelSequence] = ...,
        arguments: ArgumentsT = ...,
        with_param: Optional[Any] = ...,
        with_items: Optional[OneOrMany[Any]] = ...,
        dependencies: Optional[List[str]] = ...,
        depends: Optional[str] = ...,
    ) -> Task:
        """Create and return a Task.

        Must be invoked under a DAG context.
        """
        # Note: signature must match the Task constructor, except that while name is required for Task,
        # it is automatically inferred from the name of the decorated function when invoked.


# Pass actual class of Script to bind inputs to the ParamSpec above
def _add_type_hints(
    _script: Callable[ScriptIns, Script],
) -> Callable[
    ...,
    Callable[
        ScriptIns,  # this adds Script type hints to the underlying *library* function kwargs, i.e. `script`
        Callable[  # we will return a function that is a decorator
            [Callable[FuncIns, FuncR]],  # taking underlying *user* function
            _ScriptDecoratedFunction[  # and returning an overloaded method that can additionally return Task or Step
                FuncIns, FuncR
            ],
        ],
    ],
]:
    """Adds type hints to the decorated function."""
    return lambda func: func


@_add_type_hints(Script)
def script(**script_kwargs) -> Callable:
    """A decorator that wraps a function into a Script object.

    Using this decorator users can define a function that will be executed as a script in a container. Once the
    `Script` is returned users can use it as they generally use a `Script` e.g. as a callable inside a DAG or Steps.
    Note that invoking the function will result in the template associated with the script to be added to the
    workflow context, so users do not have to worry about that.

    Parameters
    ----------
    **script_kwargs
        Keyword arguments to be passed to the Script object.

    Returns:
    -------
    Callable
        Function that wraps a given function into a `Script`.
    """

    def script_wrapper(func: Callable[FuncIns, FuncR]) -> Callable:
        """Wraps the given callable so it can be invoked as a Step or Task.

        Parameters
        ----------
        func: Callable
            Function to wrap.

        Returns:
        -------
        Callable
            Callable that represents the `Script` object `__call__` method when in a Steps or DAG context,
            otherwise returns the callable function unchanged.
        """
        # instance methods are wrapped in `staticmethod`. Hera can capture that type and extract the underlying
        # function for remote submission since it does not depend on any class or instance attributes, so it is
        # submittable
        if isinstance(func, staticmethod):
            source: Callable = func.__func__
        else:
            source = func

        if "name" in script_kwargs:
            # take the client-provided `name` if it is submitted, pop the name for otherwise there will be two
            # kwargs called `name`
            name = script_kwargs.pop("name")
        else:
            # otherwise populate the `name` from the function name
            name = source.__name__.replace("_", "-")

        s = Script(name=name, source=source, **script_kwargs)

        @wraps(func)
        def task_wrapper(*args, **kwargs) -> Union[FuncR, Step, Task, None]:
            """Invokes a `Script` object's `__call__` method using the given SubNode (Step or Task) args/kwargs."""
            if _context.active:
                return s.__call__(*args, **kwargs)
            return func(*args, **kwargs)

        # Set the wrapped function to the original function so that we can use it later
        task_wrapper.wrapped_function = func  # type: ignore
        return task_wrapper

    return script_wrapper


class InlineScriptConstructor(ScriptConstructor):
    """`InlineScriptConstructor` is a script constructor that submits a script as a `source` to Argo.

    This script constructor is focused on taking a Python script/function "as is" for remote execution. The
    constructor processes the script to infer what parameters it needs to deserialize so the script can execute.
    The submitted script will contain prefixes such as new imports, e.g. `import os`, `import json`, etc. and
    will contain the necessary `json.loads` calls to deserialize the parameters so they are usable by the script just
    like a normal Python script/function.
    """

    add_cwd_to_sys_path: Optional[bool] = None

    @staticmethod
    def _roundtrip(source):
        tree = ast.parse(source)
        if hasattr(ast, "unparse"):
            return ast.unparse(tree)
        return ast.unparse(tree)

    def _get_param_script_portion(self, instance: Script) -> str:
        """Constructs and returns a script that loads the parameters of the specified arguments.

        Since Argo passes parameters through `{{input.parameters.name}}` it can be very cumbersome for users to
        manage that. This creates a script that automatically imports json and loads/adds code to interpret
        each independent argument into the script.

        Returns:
        -------
        str
            The string representation of the script to load.
        """
        inputs = instance._build_inputs()
        assert inputs
        extract = "import json\n"
        for param in sorted(inputs.parameters or [], key=lambda x: x.name):
            # Hera does not know what the content of the `InputFrom` is, coming from another task. In some cases
            # non-JSON encoded strings are returned, which fail the loads, but they can be used as plain strings
            # which is why this captures that in an except. This is only used for `InputFrom` cases as the extra
            # payload of the script is not necessary when regular input is set on the task via `func_params`
            if param.value_from is None:
                extract += f"""try: {param.name} = json.loads(r'''{{{{inputs.parameters.{param.name}}}}}''')\n"""
                extract += f"""except: {param.name} = r'''{{{{inputs.parameters.{param.name}}}}}'''\n"""
        return textwrap.dedent(extract) if extract != "import json\n" else ""

    def generate_source(self, instance: Script) -> str:
        """Assembles and returns a script representation of the given function.

        This also assembles any extra script material prefixed to the string source.
        The script is expected to be a callable function the client is interested in submitting
        for execution on Argo and the `script_extra` material represents the parameter loading part obtained, likely,
        through `get_param_script_portion`.

        Returns:
        -------
        str
            Final formatted script.
        """
        if not callable(instance.source):
            assert isinstance(instance.source, str)
            return instance.source
        args = inspect.getfullargspec(instance.source).args
        script = ""
        # Argo will save the script as a file and run it with cmd:
        # - python /argo/staging/script
        # However, this prevents the script from importing modules in its cwd,
        # since it's looking for files relative to the script path.
        # We fix this by appending the cwd path to sys:
        if instance.add_cwd_to_sys_path or self.add_cwd_to_sys_path:
            script = "import os\nimport sys\nsys.path.append(os.getcwd())\n"

        script_extra = self._get_param_script_portion(instance) if args else None
        if script_extra:
            script += copy.deepcopy(script_extra)
            script += "\n"

        # We use ast parse/unparse to get the source code of the function
        # in order to have consistent looking functions and getting rid of any comments
        # parsing issues.
        # See https://github.com/argoproj-labs/hera/issues/572
        content = self._roundtrip(textwrap.dedent(inspect.getsource(instance.source))).splitlines()
        for i, line in enumerate(content):
            if line.startswith("def") or line.startswith("async def"):
                break

        s = "\n".join(content[i + 1 :])
        script += textwrap.dedent(s)
        return textwrap.dedent(script)


class RunnerScriptConstructor(ScriptConstructor):
    """`RunnerScriptConstructor` is a script constructor that runs a script in a container.

    The runner script, also known as "The Hera runner", takes a script/Python function definition, infers the path
    to the function (module import), assembles a path to invoke the function, and passes any specified parameters
    to the function. This helps users "save" on the `source` space required for submitting a function for remote
    execution on Argo. Execution within the container *requires* the executing container to include the file that
    contains the submitted script. More specifically, the container must be created in some process (e.g. CI), so that
    it conains the script to run remotely.
    """

    outputs_directory: Optional[str] = None
    """Used for saving outputs when defined using annotations."""

    volume_for_outputs: Optional[_BaseVolume] = None
    """Volume to use if saving outputs when defined using annotations."""

    DEFAULT_HERA_OUTPUTS_DIRECTORY: str = "/tmp/hera-outputs"
    """Used as the default value for when the outputs_directory is not set"""

    pydantic_mode: Optional[Literal[1, 2]] = None
    """Used for selecting the pydantic version used for BaseModels.
    Allows for using pydantic.v1 BaseModels with pydantic v2.
    Defaults to the installed version of Pydantic."""

    @validator("pydantic_mode", always=True)
    def _pydantic_mode(cls, value: Optional[Literal[1, 2]]) -> Optional[Literal[1, 2]]:
        if value and value > _PYDANTIC_VERSION:
            raise ValueError("v2 pydantic mode only available for pydantic>=2")
        return value

    def transform_values(self, cls: Type[Script], values: Any) -> Any:
        """A function that can inspect the Script instance and generate the source field."""
        if not callable(values.get("source")):
            return values

        if values.get("args") is not None:
            raise ValueError("Cannot specify args when callable is True")
        values["args"] = [
            "-m",
            "hera.workflows.runner",
            "-e",
            f"{values['source'].__module__}:{values['source'].__name__}",
        ]

        return values

    def generate_source(self, instance: Script) -> str:
        """A function that can inspect the Script instance and generate the source field."""
        return f"{g.inputs.parameters:$}"

    def transform_script_template_post_build(
        self, instance: "Script", script: _ModelScriptTemplate
    ) -> _ModelScriptTemplate:
        """A hook to transform the generated script template."""
        script_env = []

        if self.outputs_directory:
            script_env.append(EnvVar(name="hera__outputs_directory", value=self.outputs_directory))
        if self.pydantic_mode:
            script_env.append(EnvVar(name="hera__pydantic_mode", value=str(self.pydantic_mode)))
        if _flag_enabled(_SCRIPT_PYDANTIC_IO_FLAG):
            script_env.append(EnvVar(name="hera__script_pydantic_io", value=""))

        if script_env:
            if not script.env:
                # If user did not set any env vars themselves then we need to initialise the list
                script.env = []

            script.env.extend(script_env)

        return script


__all__ = ["Script", "script", "ScriptConstructor", "InlineScriptConstructor", "RunnerScriptConstructor"]



# generated by datamodel-codegen:
#   filename:  argo-workflows-3.6.2.json

from __future__ import annotations

from datetime import datetime
from typing import Annotated, Dict, List, Optional

from hera.shared._pydantic import BaseModel, Field


class Time(BaseModel):
    __root__: Annotated[
        datetime,
        Field(
            description=(
                "Time is a wrapper around time.Time which supports correct marshaling"
                " to YAML and JSON.  Wrappers are provided for many of the factory"
                " methods that the time package offers."
            )
        ),
    ]


class ListMeta(BaseModel):
    continue_: Annotated[
        Optional[str],
        Field(
            alias="continue",
            description=(
                "continue may be set if the user set a limit on the number of items"
                " returned, and indicates that the server has more data available. The"
                " value is opaque and may be used to issue another request to the"
                " endpoint that served this list to retrieve the next set of available"
                " objects. Continuing a consistent list may not be possible if the"
                " server configuration has changed or more than a few minutes have"
                " passed. The resourceVersion field returned when using this continue"
                " value will be identical to the value in the first response, unless"
                " you have received this token from an error message."
            ),
        ),
    ] = None
    remaining_item_count: Annotated[
        Optional[int],
        Field(
            alias="remainingItemCount",
            description=(
                "remainingItemCount is the number of subsequent items in the list which"
                " are not included in this list response. If the list request contained"
                " label or field selectors, then the number of remaining items is"
                " unknown and the field will be left unset and omitted during"
                " serialization. If the list is complete (either because it is not"
                " chunking or because this is the last chunk), then there are no more"
                " remaining items and this field will be left unset and omitted during"
                " serialization. Servers older than v1.15 do not set this field. The"
                " intended use of the remainingItemCount is *estimating* the size of a"
                " collection. Clients should not rely on the remainingItemCount to be"
                " set or to be exact."
            ),
        ),
    ] = None
    resource_version: Annotated[
        Optional[str],
        Field(
            alias="resourceVersion",
            description=(
                "String that identifies the server's internal version of this object"
                " that can be used by clients to determine when objects have changed."
                " Value must be treated as opaque by clients and passed unmodified back"
                " to the server. Populated by the system. Read-only. More info:"
                " https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#concurrency-control-and-consistency"
            ),
        ),
    ] = None
    self_link: Annotated[
        Optional[str],
        Field(
            alias="selfLink",
            description=(
                "Deprecated: selfLink is a legacy read-only field that is no longer populated by the system."
            ),
        ),
    ] = None


class GroupVersionResource(BaseModel):
    group: Optional[str] = None
    resource: Optional[str] = None
    version: Optional[str] = None


class CreateOptions(BaseModel):
    dry_run: Annotated[
        Optional[List[str]],
        Field(
            alias="dryRun",
            title=(
                "When present, indicates that modifications should not be\npersisted."
                " An invalid or unrecognized dryRun directive will\nresult in an error"
                " response and no further processing of the\nrequest. Valid values"
                " are:\n- All: all dry run stages will be"
                " processed\n+optional\n+listType=atomic"
            ),
        ),
    ] = None
    field_manager: Annotated[
        Optional[str],
        Field(
            alias="fieldManager",
            title=(
                "fieldManager is a name associated with the actor or entity\nthat is"
                " making these changes. The value must be less than or\n128 characters"
                " long, and only contain printable characters,\nas defined by"
                " https://golang.org/pkg/unicode/#IsPrint.\n+optional"
            ),
        ),
    ] = None
    field_validation: Annotated[
        Optional[str],
        Field(
            alias="fieldValidation",
            title=(
                "fieldValidation instructs the server on how to handle\nobjects in the"
                " request (POST/PUT/PATCH) containing unknown\nor duplicate fields."
                " Valid values are:\n- Ignore: This will ignore any unknown fields that"
                " are silently\ndropped from the object, and will ignore all but the"
                " last duplicate\nfield that the decoder encounters. This is the"
                " default behavior\nprior to v1.23.\n- Warn: This will send a warning"
                " via the standard warning response\nheader for each unknown field that"
                " is dropped from the object, and\nfor each duplicate field that is"
                " encountered. The request will\nstill succeed if there are no other"
                " errors, and will only persist\nthe last of any duplicate fields. This"
                " is the default in v1.23+\n- Strict: This will fail the request with a"
                " BadRequest error if\nany unknown fields would be dropped from the"
                " object, or if any\nduplicate fields are present. The error returned"
                " from the server\nwill contain all unknown and duplicate fields"
                " encountered.\n+optional"
            ),
        ),
    ] = None


class OwnerReference(BaseModel):
    api_version: Annotated[str, Field(alias="apiVersion", description="API version of the referent.")]
    block_owner_deletion: Annotated[
        Optional[bool],
        Field(
            alias="blockOwnerDeletion",
            description=(
                'If true, AND if the owner has the "foregroundDeletion" finalizer, then'
                " the owner cannot be deleted from the key-value store until this"
                " reference is removed. See"
                " https://kubernetes.io/docs/concepts/architecture/garbage-collection/#foreground-deletion"
                " for how the garbage collector interacts with this field and enforces"
                " the foreground deletion. Defaults to false. To set this field, a user"
                ' needs "delete" permission of the owner, otherwise 422 (Unprocessable'
                " Entity) will be returned."
            ),
        ),
    ] = None
    controller: Annotated[
        Optional[bool],
        Field(description="If true, this reference points to the managing controller."),
    ] = None
    kind: Annotated[
        str,
        Field(
            description=(
                "Kind of the referent. More info:"
                " https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds"
            )
        ),
    ]
    name: Annotated[
        str,
        Field(
            description=(
                "Name of the referent. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/names#names"
            )
        ),
    ]
    uid: Annotated[
        str,
        Field(
            description=(
                "UID of the referent. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/names#uids"
            )
        ),
    ]


class MicroTime(BaseModel):
    __root__: Annotated[
        datetime,
        Field(description="MicroTime is version of Time with microsecond level precision."),
    ]


class FieldsV1(BaseModel):
    pass


class LabelSelectorRequirement(BaseModel):
    key: Annotated[str, Field(description="key is the label key that the selector applies to.")]
    operator: Annotated[
        str,
        Field(
            description=(
                "operator represents a key's relationship to a set of values. Valid"
                " operators are In, NotIn, Exists and DoesNotExist."
            )
        ),
    ]
    values: Annotated[
        Optional[List[str]],
        Field(
            description=(
                "values is an array of string values. If the operator is In or NotIn,"
                " the values array must be non-empty. If the operator is Exists or"
                " DoesNotExist, the values array must be empty. This array is replaced"
                " during a strategic merge patch."
            )
        ),
    ] = None


class ManagedFieldsEntry(BaseModel):
    api_version: Annotated[
        Optional[str],
        Field(
            alias="apiVersion",
            description=(
                "APIVersion defines the version of this resource that this field set"
                ' applies to. The format is "group/version" just like the top-level'
                " APIVersion field. It is necessary to track the version of a field set"
                " because it cannot be automatically converted."
            ),
        ),
    ] = None
    fields_type: Annotated[
        Optional[str],
        Field(
            alias="fieldsType",
            description=(
                "FieldsType is the discriminator for the different fields format and"
                ' version. There is currently only one possible value: "FieldsV1"'
            ),
        ),
    ] = None
    fields_v1: Annotated[
        Optional[FieldsV1],
        Field(
            alias="fieldsV1",
            description=('FieldsV1 holds the first JSON version format as described in the "FieldsV1" type.'),
        ),
    ] = None
    manager: Annotated[
        Optional[str],
        Field(description=("Manager is an identifier of the workflow managing these fields.")),
    ] = None
    operation: Annotated[
        Optional[str],
        Field(
            description=(
                "Operation is the type of operation which lead to this"
                " ManagedFieldsEntry being created. The only valid values for this"
                " field are 'Apply' and 'Update'."
            )
        ),
    ] = None
    subresource: Annotated[
        Optional[str],
        Field(
            description=(
                "Subresource is the name of the subresource used to update that object,"
                " or empty string if the object was updated through the main resource."
                " The value of this field is used to distinguish between managers, even"
                " if they share the same name. For example, a status update will be"
                " distinct from a regular update using the same manager name. Note that"
                " the APIVersion field is not related to the Subresource field and it"
                " always corresponds to the version of the main resource."
            )
        ),
    ] = None
    time: Annotated[
        Optional[Time],
        Field(
            description=(
                "Time is the timestamp of when the ManagedFields entry was added. The"
                " timestamp will also be updated if a field is added, the manager"
                " changes any of the owned fields value or removes a field. The"
                " timestamp does not update when a field is removed from the entry"
                " because another manager took it over."
            )
        ),
    ] = None


class LabelSelector(BaseModel):
    match_expressions: Annotated[
        Optional[List[LabelSelectorRequirement]],
        Field(
            alias="matchExpressions",
            description=("matchExpressions is a list of label selector requirements. The requirements are ANDed."),
        ),
    ] = None
    match_labels: Annotated[
        Optional[Dict[str, str]],
        Field(
            alias="matchLabels",
            description=(
                "matchLabels is a map of {key,value} pairs. A single {key,value} in the"
                " matchLabels map is equivalent to an element of matchExpressions,"
                ' whose key field is "key", the operator is "In", and the values array'
                ' contains only "value". The requirements are ANDed.'
            ),
        ),
    ] = None


class ObjectMeta(BaseModel):
    annotations: Annotated[
        Optional[Dict[str, str]],
        Field(
            description=(
                "Annotations is an unstructured key value map stored with a resource"
                " that may be set by external tools to store and retrieve arbitrary"
                " metadata. They are not queryable and should be preserved when"
                " modifying objects. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/annotations"
            )
        ),
    ] = None
    creation_timestamp: Annotated[
        Optional[Time],
        Field(
            alias="creationTimestamp",
            description=(
                "CreationTimestamp is a timestamp representing the server time when"
                " this object was created. It is not guaranteed to be set in"
                " happens-before order across separate operations. Clients may not set"
                " this value. It is represented in RFC3339 form and is in"
                " UTC.\n\nPopulated by the system. Read-only. Null for lists. More"
                " info:"
                " https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata"
            ),
        ),
    ] = None
    deletion_grace_period_seconds: Annotated[
        Optional[int],
        Field(
            alias="deletionGracePeriodSeconds",
            description=(
                "Number of seconds allowed for this object to gracefully terminate"
                " before it will be removed from the system. Only set when"
                " deletionTimestamp is also set. May only be shortened. Read-only."
            ),
        ),
    ] = None
    deletion_timestamp: Annotated[
        Optional[Time],
        Field(
            alias="deletionTimestamp",
            description=(
                "DeletionTimestamp is RFC 3339 date and time at which this resource"
                " will be deleted. This field is set by the server when a graceful"
                " deletion is requested by the user, and is not directly settable by a"
                " client. The resource is expected to be deleted (no longer visible"
                " from resource lists, and not reachable by name) after the time in"
                " this field, once the finalizers list is empty. As long as the"
                " finalizers list contains items, deletion is blocked. Once the"
                " deletionTimestamp is set, this value may not be unset or be set"
                " further into the future, although it may be shortened or the resource"
                " may be deleted prior to this time. For example, a user may request"
                " that a pod is deleted in 30 seconds. The Kubelet will react by"
                " sending a graceful termination signal to the containers in the pod."
                " After that 30 seconds, the Kubelet will send a hard termination"
                " signal (SIGKILL) to the container and after cleanup, remove the pod"
                " from the API. In the presence of network partitions, this object may"
                " still exist after this timestamp, until an administrator or automated"
                " process can determine the resource is fully terminated. If not set,"
                " graceful deletion of the object has not been requested.\n\nPopulated"
                " by the system when a graceful deletion is requested. Read-only. More"
                " info:"
                " https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata"
            ),
        ),
    ] = None
    finalizers: Annotated[
        Optional[List[str]],
        Field(
            description=(
                "Must be empty before the object is deleted from the registry. Each"
                " entry is an identifier for the responsible component that will remove"
                " the entry from the list. If the deletionTimestamp of the object is"
                " non-nil, entries in this list can only be removed. Finalizers may be"
                " processed and removed in any order.  Order is NOT enforced because it"
                " introduces significant risk of stuck finalizers. finalizers is a"
                " shared field, any actor with permission can reorder it. If the"
                " finalizer list is processed in order, then this can lead to a"
                " situation in which the component responsible for the first finalizer"
                " in the list is waiting for a signal (field value, external system, or"
                " other) produced by a component responsible for a finalizer later in"
                " the list, resulting in a deadlock. Without enforced ordering"
                " finalizers are free to order amongst themselves and are not"
                " vulnerable to ordering changes in the list."
            )
        ),
    ] = None
    generate_name: Annotated[
        Optional[str],
        Field(
            alias="generateName",
            description=(
                "GenerateName is an optional prefix, used by the server, to generate a"
                " unique name ONLY IF the Name field has not been provided. If this"
                " field is used, the name returned to the client will be different than"
                " the name passed. This value will also be combined with a unique"
                " suffix. The provided value has the same validation rules as the Name"
                " field, and may be truncated by the length of the suffix required to"
                " make the value unique on the server.\n\nIf this field is specified"
                " and the generated name exists, the server will return a"
                " 409.\n\nApplied only if Name is not specified. More info:"
                " https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#idempotency"
            ),
        ),
    ] = None
    generation: Annotated[
        Optional[int],
        Field(
            description=(
                "A sequence number representing a specific generation of the desired"
                " state. Populated by the system. Read-only."
            )
        ),
    ] = None
    labels: Annotated[
        Optional[Dict[str, str]],
        Field(
            description=(
                "Map of string keys and values that can be used to organize and"
                " categorize (scope and select) objects. May match selectors of"
                " replication controllers and services. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/labels"
            )
        ),
    ] = None
    managed_fields: Annotated[
        Optional[List[ManagedFieldsEntry]],
        Field(
            alias="managedFields",
            description=(
                "ManagedFields maps workflow-id and version to the set of fields that"
                " are managed by that workflow. This is mostly for internal"
                " housekeeping, and users typically shouldn't need to set or"
                " understand this field. A workflow can be the user's name, a"
                " controller's name, or the name of a specific apply path like"
                ' "ci-cd". The set of fields is always in the version that the workflow'
                " used when modifying the object."
            ),
        ),
    ] = None
    name: Annotated[
        Optional[str],
        Field(
            description=(
                "Name must be unique within a namespace. Is required when creating"
                " resources, although some resources may allow a client to request the"
                " generation of an appropriate name automatically. Name is primarily"
                " intended for creation idempotence and configuration definition."
                " Cannot be updated. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/names#names"
            )
        ),
    ] = None
    namespace: Annotated[
        Optional[str],
        Field(
            description=(
                "Namespace defines the space within which each name must be unique. An"
                ' empty namespace is equivalent to the "default" namespace, but'
                ' "default" is the canonical representation. Not all objects are'
                " required to be scoped to a namespace - the value of this field for"
                " those objects will be empty.\n\nMust be a DNS_LABEL. Cannot be"
                " updated. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/namespaces"
            )
        ),
    ] = None
    owner_references: Annotated[
        Optional[List[OwnerReference]],
        Field(
            alias="ownerReferences",
            description=(
                "List of objects depended by this object. If ALL objects in the list"
                " have been deleted, this object will be garbage collected. If this"
                " object is managed by a controller, then an entry in this list will"
                " point to this controller, with the controller field set to true."
                " There cannot be more than one managing controller."
            ),
        ),
    ] = None
    resource_version: Annotated[
        Optional[str],
        Field(
            alias="resourceVersion",
            description=(
                "An opaque value that represents the internal version of this object"
                " that can be used by clients to determine when objects have changed."
                " May be used for optimistic concurrency, change detection, and the"
                " watch operation on a resource or set of resources. Clients must treat"
                " these values as opaque and passed unmodified back to the server. They"
                " may only be valid for a particular resource or set of"
                " resources.\n\nPopulated by the system. Read-only. Value must be"
                " treated as opaque by clients and . More info:"
                " https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#concurrency-control-and-consistency"
            ),
        ),
    ] = None
    self_link: Annotated[
        Optional[str],
        Field(
            alias="selfLink",
            description=(
                "Deprecated: selfLink is a legacy read-only field that is no longer populated by the system."
            ),
        ),
    ] = None
    uid: Annotated[
        Optional[str],
        Field(
            description=(
                "UID is the unique in time and space value for this object. It is"
                " typically generated by the server on successful creation of a"
                " resource and is not allowed to change on PUT operations.\n\nPopulated"
                " by the system. Read-only. More info:"
                " https://kubernetes.io/docs/concepts/overview/working-with-objects/names#uids"
            )
        ),
    ] = None



# generated by datamodel-codegen:
#   filename:  argo-workflows-3.6.2.json

from __future__ import annotations

from typing import Annotated, Optional

from hera.shared._pydantic import BaseModel, Field

from .io.argoproj.events import v1alpha1
from .io.k8s.apimachinery.pkg.apis.meta import v1


class EventSourceDeletedResponse(BaseModel):
    pass


class LogEntry(BaseModel):
    event_name: Annotated[
        Optional[str],
        Field(alias="eventName", title="optional - the event name (e.g. `example`)"),
    ] = None
    event_source_name: Annotated[Optional[str], Field(alias="eventSourceName")] = None
    event_source_type: Annotated[
        Optional[str],
        Field(
            alias="eventSourceType",
            title="optional - the event source type (e.g. `webhook`)",
        ),
    ] = None
    level: Optional[str] = None
    msg: Optional[str] = None
    namespace: Optional[str] = None
    time: Optional[v1.Time] = None


class CreateEventSourceRequest(BaseModel):
    event_source: Annotated[Optional[v1alpha1.EventSource], Field(alias="eventSource")] = None
    namespace: Optional[str] = None


class EventSourceWatchEvent(BaseModel):
    object: Optional[v1alpha1.EventSource] = None
    type: Optional[str] = None


class UpdateEventSourceRequest(BaseModel):
    event_source: Annotated[Optional[v1alpha1.EventSource], Field(alias="eventSource")] = None
    name: Optional[str] = None
    namespace: Optional[str] = None



"""The util module contains the functionality required for the script runner."""

import argparse
import functools
import importlib
import inspect
import json
import os
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, cast

if sys.version_info >= (3, 10):
    from types import NoneType
else:
    NoneType = type(None)

from hera.shared._pydantic import _PYDANTIC_VERSION
from hera.shared._type_util import (
    get_workflow_annotation,
    origin_type_issubtype,
    unwrap_annotation,
)
from hera.shared.serialization import serialize
from hera.workflows import Artifact
from hera.workflows._runner.script_annotations_util import (
    _map_argo_inputs_to_function,
    _save_annotated_return_outputs,
    _save_dummy_outputs,
)
from hera.workflows.artifact import ArtifactLoader
from hera.workflows.io.v1 import (
    Output as OutputV1,
)
from hera.workflows.script import _extract_return_annotation_output

if _PYDANTIC_VERSION == 2:
    from pydantic.type_adapter import TypeAdapter  # type: ignore
    from pydantic.v1 import parse_obj_as  # type: ignore

    from hera.workflows.io.v2 import (  # type: ignore
        Output as OutputV2,
    )
else:
    from pydantic import parse_obj_as

    from hera.workflows.io.v1 import (  # type: ignore
        Output as OutputV2,
    )


def _ignore_unmatched_kwargs(f: Callable) -> Callable:
    """Make function ignore unmatched kwargs.

    If the function already has the catch all **kwargs, do nothing.
    """
    if _contains_var_kwarg(f):
        return f

    @functools.wraps(f)
    def inner(**kwargs):
        # filter out kwargs that are not part of the function signature
        # and transform them to the correct type
        filtered_kwargs = {key: _parse(value, key, f) for key, value in kwargs.items() if _is_kwarg_of(key, f)}
        return f(**filtered_kwargs)

    return inner


def _contains_var_kwarg(f: Callable) -> bool:
    """Tells whether the given callable contains a keyword argument."""
    return any(param.kind == inspect.Parameter.VAR_KEYWORD for param in inspect.signature(f).parameters.values())


def _is_kwarg_of(key: str, f: Callable) -> bool:
    """Tells whether the given `key` identifies a keyword argument of the given callable."""
    param = inspect.signature(f).parameters.get(key)
    return param is not None and (
        param.kind is inspect.Parameter.KEYWORD_ONLY or param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD
    )


def _parse(value: str, key: str, f: Callable) -> Any:
    """Parse a value to the correct type.

    Args:
        value: The value to parse.
        key: The name of the kwarg.
        f: The function to parse the value for.

    Returns:
        The parsed value.

    """
    if _is_str_kwarg_of(key, f) or _is_artifact_loaded(key, f) or _is_output_kwarg(key, f):
        return value
    try:
        type_ = _get_unannotated_type(key, f)
        loaded_json_value = json.loads(value)

        if not type_:
            return loaded_json_value

        _pydantic_mode = int(os.environ.get("hera__pydantic_mode", _PYDANTIC_VERSION))
        if _pydantic_mode == 1:
            return parse_obj_as(type_, loaded_json_value)
        else:
            return TypeAdapter(type_).validate_python(loaded_json_value)
    except (json.JSONDecodeError, TypeError):
        return value


def _get_function_param_annotation(key: str, f: Callable) -> Optional[type]:
    func_param_annotation = inspect.signature(f).parameters[key].annotation
    if func_param_annotation is inspect.Parameter.empty:
        return None
    return func_param_annotation


def _get_unannotated_type(key: str, f: Callable) -> Optional[type]:
    """Get the type of function param without the 'Annotated' outer type."""
    type_ = _get_function_param_annotation(key, f)
    if type_ is None:
        return None
    return unwrap_annotation(type_)


def _is_str_kwarg_of(key: str, f: Callable) -> bool:
    """Check if param `key` of function `f` has a type annotation that can be interpreted as a subclass of str."""
    if func_param_annotation := _get_function_param_annotation(key, f):
        return origin_type_issubtype(func_param_annotation, (str, NoneType))
    return False


def _is_artifact_loaded(key: str, f: Callable) -> bool:
    """Check if param `key` of function `f` is actually an Artifact that has already been loaded."""
    if param_annotation := _get_function_param_annotation(key, f):
        if (artifact := get_workflow_annotation(param_annotation)) and isinstance(artifact, Artifact):
            return artifact.loader == ArtifactLoader.json.value
    return False


def _is_output_kwarg(key: str, f: Callable) -> bool:
    """Check if param `key` of function `f` is an output Artifact/Parameter."""
    if param_annotation := _get_function_param_annotation(key, f):
        if param_or_artifact := get_workflow_annotation(param_annotation):
            return bool(param_or_artifact.output)
    return False


def _runner(entrypoint: str, kwargs_list: List) -> Any:
    """Run the function defined by the entrypoint with the given list of kwargs.

    Args:
        entrypoint: The module path to the script within the container to execute. "package.submodule:function"
        kwargs_list: A list of dicts with "name" and "value" keys, representing the kwargs of the function.

    Returns:
        The result of the function or `None` if the outputs are to be saved.
    """
    # import the module and get the function
    module, function_name = entrypoint.split(":")
    function: Callable = getattr(importlib.import_module(module), function_name)
    # if the function is wrapped, unwrap it
    # this may happen if the function is decorated with @script
    if hasattr(function, "wrapped_function"):
        function = function.wrapped_function
    # convert the kwargs list to a dict
    kwargs: Dict[str, str] = {}
    for kwarg in kwargs_list:
        if "name" not in kwarg or "value" not in kwarg:
            continue
        # sanitize the key for python
        key = cast(str, serialize(kwarg["name"]))
        value = kwarg["value"]
        kwargs[key] = value

    kwargs = _map_argo_inputs_to_function(function, kwargs)

    # The imported validate_arguments uses smart union by default just in case clients do not rely on it. This means that if a function uses a union
    # type for any of its inputs, then this will at least try to map those types correctly if the input object is
    # not a pydantic model with smart_union enabled
    _pydantic_mode = int(os.environ.get("hera__pydantic_mode", _PYDANTIC_VERSION))
    if _pydantic_mode == 2:
        from pydantic import validate_call  # type: ignore

        function = validate_call(function)
    else:
        if _PYDANTIC_VERSION == 1:
            from pydantic import validate_arguments
        else:
            from pydantic.v1 import validate_arguments  # type: ignore
        function = validate_arguments(function, config=dict(smart_union=True, arbitrary_types_allowed=True))  # type: ignore

    function = _ignore_unmatched_kwargs(function)

    output_annotations = _extract_return_annotation_output(function)

    if output_annotations:
        # This will save outputs returned from the function only. Any function parameters/artifacts marked as
        # outputs should be written to within the function itself.
        try:
            output = _save_annotated_return_outputs(function(**kwargs), output_annotations)
        except Exception as e:
            _save_dummy_outputs(output_annotations)
            raise e
        return output or None

    return function(**kwargs)


def _parse_args() -> argparse.Namespace:
    """Creates an argparse for the runner function.

    The returned argparse takes a module and function name as flags and a path to a json file as an argument.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--entrypoint", "-e", type=str, required=True)
    parser.add_argument("args_path", type=Path)
    return parser.parse_args()


def _run() -> None:
    """Runs a function from a specific path using parsed arguments from Argo.

    Note that this prints the result of the function to stdout, which is the normal mode of operation for Argo. Any
    output of a Python function submitted via a `Script.source` field results in outputs sent to stdout.
    """
    args = _parse_args()
    # 1. Protect against trying to json.loads on empty files with inner `or r"[]`
    # 2. Protect against files containing `null` as text with outer `or []` (as a result of using
    #    `{{inputs.parameters}}` where the parameters key doesn't exist in `inputs`)
    kwargs_list = json.loads(args.args_path.read_text() or r"[]") or []
    assert isinstance(kwargs_list, List)
    result = _runner(args.entrypoint, kwargs_list)
    if not result:
        return

    if isinstance(result, (OutputV1, OutputV2)):
        print(serialize(result.result))
        exit(result.exit_code)

    print(serialize(result))



"""The dag module provides the DAG class.

See https://argoproj.github.io/argo-workflows/walk-through/dag/
for more on DAGs (Directed Acyclic Graphs).
"""

from __future__ import annotations

from typing import Any, List, Optional, Set, Union

from hera.shared._pydantic import PrivateAttr
from hera.workflows._meta_mixins import CallableTemplateMixin, ContextMixin
from hera.workflows._mixins import IOMixin, TemplateMixin
from hera.workflows.exceptions import InvalidType, NodeNameConflict
from hera.workflows.models import (
    DAGTask,
    DAGTemplate as _ModelDAGTemplate,
    Template as _ModelTemplate,
)
from hera.workflows.task import Task


class DAG(
    IOMixin,
    TemplateMixin,
    CallableTemplateMixin,
    ContextMixin,
):
    """A DAG template invocator is used to define Task dependencies as an acyclic graph.

    DAG implements the contextmanager interface so allows usage of `with`, under which any
    `hera.workflows.task.Task` objects instantiated will be added to the DAG's list of Tasks.

    Examples:
        >>> @script()
        >>> def foo() -> None:
        >>>     print(42)
        >>>
        >>> with DAG(...) as dag:
        >>>     foo()
    """

    fail_fast: Optional[bool] = None
    target: Optional[str] = None
    tasks: List[Union[Task, DAGTask]] = []

    _node_names = PrivateAttr(default_factory=set)
    _current_task_depends: Set[str] = PrivateAttr(set())

    def _add_sub(self, node: Any):
        if not isinstance(node, Task):
            raise InvalidType(type(node))
        if node.name in self._node_names:
            raise NodeNameConflict(f"Found multiple Task nodes with name: {node.name}")
        self._node_names.add(node.name)
        self.tasks.append(node)

    def _build_template(self) -> _ModelTemplate:
        """Builds the auto-generated `Template` representation of the `DAG`."""
        tasks = []
        for task in self.tasks:
            if isinstance(task, Task):
                tasks.append(task._build_dag_task())
            else:
                tasks.append(task)
        return _ModelTemplate(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            archive_location=self.archive_location,
            automount_service_account_token=self.automount_service_account_token,
            daemon=self.daemon,
            dag=_ModelDAGTemplate(fail_fast=self.fail_fast, target=self.target, tasks=tasks),
            executor=self.executor,
            fail_fast=self.fail_fast,
            host_aliases=self.host_aliases,
            init_containers=self._build_init_containers(),
            inputs=self._build_inputs(),
            memoize=self.memoize,
            metadata=self._build_metadata(),
            metrics=self._build_metrics(),
            name=self.name,
            node_selector=self.node_selector,
            outputs=self._build_outputs(),
            parallelism=self.parallelism,
            plugin=self.plugin,
            pod_spec_patch=self.pod_spec_patch,
            priority=self.priority,
            priority_class_name=self.priority_class_name,
            retry_strategy=self.retry_strategy,
            scheduler_name=self.scheduler_name,
            security_context=self.pod_security_context,
            service_account_name=self.service_account_name,
            sidecars=self._build_sidecars(),
            synchronization=self.synchronization,
            timeout=self.timeout,
            tolerations=self.tolerations,
        )


__all__ = ["DAG"]



"""The exceptions module provides exception types required for the Hera workflows package."""


class WorkflowsException(Exception):
    """Base Hera workflows exception."""

    ...


class InvalidType(WorkflowsException):
    """Exception raised when an invalid type is submitted to a Hera object's field or functionality."""

    ...


class InvalidTemplateCall(WorkflowsException):
    """Exception raised when an invalid template call is performed."""

    ...


class InvalidDispatchType(WorkflowsException):
    """Exception raised when Hera attempts to dispatch a hook and it fails to do so."""

    ...


class TemplateNameConflict(WorkflowsException):
    """Currently unused."""

    ...


class NodeNameConflict(WorkflowsException):
    """Exception raised when multiple Task/Step are found with the same name."""

    ...


__all__ = ["InvalidType", "InvalidTemplateCall", "InvalidDispatchType", "TemplateNameConflict", "NodeNameConflict"]



"""The user container module provides user container functionality and objects."""

from typing import List, Optional, Union, cast

from hera.workflows.env import _BaseEnv
from hera.workflows.env_from import _BaseEnvFrom
from hera.workflows.models import (
    EnvFromSource,
    EnvVar,
    ImagePullPolicy,
    ResourceRequirements,
    UserContainer as _ModelUserContainer,
    VolumeMount as _ModelVolumeMount,
)
from hera.workflows.resources import Resources
from hera.workflows.volume import _BaseVolume


class UserContainer(_ModelUserContainer):
    """`UserContainer` is a container type that is specifically used as a side container."""

    # TODO: Use EnvMixin (currently a circular import)
    env: Optional[List[Union[_BaseEnv, EnvVar]]] = None  # type: ignore[assignment]
    env_from: Optional[List[Union[_BaseEnvFrom, EnvFromSource]]] = None  # type: ignore[assignment]
    image_pull_policy: Optional[Union[str, ImagePullPolicy]] = None  # type: ignore[assignment]
    resources: Optional[Union[Resources, ResourceRequirements]] = None  # type: ignore[assignment]
    volumes: Optional[List[_BaseVolume]] = None

    def _build_image_pull_policy(self) -> Optional[str]:
        """Processes the image pull policy field and optionally returns a string value from `ImagePullPolicy` enum."""
        if self.image_pull_policy is None:
            return None
        elif isinstance(self.image_pull_policy, ImagePullPolicy):
            return self.image_pull_policy.value

        # this helps map image pull policy values as a convenience
        policy_mapper = {
            # the following 2 are "normal" entries
            **{ipp.name: ipp for ipp in ImagePullPolicy},
            **{ipp.value: ipp for ipp in ImagePullPolicy},
            # some users might submit the policy in lowercase & without underscores
            **{ipp.value.lower(): ipp for ipp in ImagePullPolicy},
        }
        try:
            return ImagePullPolicy[policy_mapper[self.image_pull_policy].name].value
        except KeyError as e:
            raise KeyError(
                f"Supplied image policy {self.image_pull_policy} is not valid. "
                "Use one of {ImagePullPolicy.__members__}"
            ) from e

    def _build_volume_mounts(self) -> Optional[List[_ModelVolumeMount]]:
        """Processes the volume mounts field and returns a generated `VolumeMount`."""
        volume_mounts = self.volume_mounts
        # extra volume mounts stem from using custom Hera volumes, which dynamically provision PVs + PVCs
        extra_volume_mounts = None if self.volumes is None else [v._build_volume_mount() for v in self.volumes]
        if volume_mounts is None:
            volume_mounts = extra_volume_mounts
        elif extra_volume_mounts is not None:
            volume_mounts.extend(extra_volume_mounts)

        return volume_mounts

    def build(self) -> _ModelUserContainer:
        """Builds the Hera auto-generated model of the user container."""
        env: List[EnvVar] = [
            var if isinstance(var, EnvVar) else cast(_BaseEnv, var).build() for var in (self.env or [])
        ]
        env_from: List[EnvFromSource] = [
            var if isinstance(var, EnvFromSource) else cast(_BaseEnvFrom, var).build() for var in (self.env_from or [])
        ]

        return _ModelUserContainer(
            args=self.args,
            command=self.command,
            env=env or None,
            env_from=env_from or None,
            image=self.image,
            image_pull_policy=self._build_image_pull_policy(),
            lifecycle=self.lifecycle,
            liveness_probe=self.liveness_probe,
            mirror_volume_mounts=self.mirror_volume_mounts,
            name=self.name,
            ports=self.ports,
            readiness_probe=self.readiness_probe,
            resources=self.resources
            if isinstance(self.resources, ResourceRequirements)
            else self.resources.build()
            if self.resources
            else None,
            security_context=self.security_context,
            startup_probe=self.startup_probe,
            stdin=self.stdin,
            stdin_once=self.stdin_once,
            termination_message_path=self.termination_message_path,
            termination_message_policy=self.termination_message_policy,
            tty=self.tty,
            volume_devices=self.volume_devices,
            volume_mounts=self._build_volume_mounts(),
            working_dir=self.working_dir,
        )


__all__ = ["UserContainer"]



"""Hera IO models."""

from pydantic import VERSION

if VERSION.split(".")[0] == "2":
    from hera.workflows.io.v2 import Input, Output

else:
    from hera.workflows.io.v1 import Input, Output  # type: ignore


__all__ = [
    "Input",
    "Output",
]



"""Data module provides Hera objects for Argo data functionality, such as sourcing data + applying transformations."""

from typing import List, Union

from hera.expr._node import Node
from hera.workflows import models as m
from hera.workflows._meta_mixins import CallableTemplateMixin
from hera.workflows._mixins import IOMixin, TemplateMixin
from hera.workflows.artifact import Artifact


class Data(TemplateMixin, IOMixin, CallableTemplateMixin):
    """`Data` implements the Argo data template representation.

    Data can be used to indicate that some data, identified by a `source`, should be processed via the specified
    `transformations`. The `transformations` field can be either expressed via a pure `str` or via a `hera.expr`,
    which transpiles the expression into a statement that can be processed by Argo.
    """

    source: Union[m.DataSource, m.ArtifactPaths, Artifact]
    transformations: List[Union[str, Node]] = []

    def _build_source(self) -> m.DataSource:
        """Builds the generated `DataSource`."""
        if isinstance(self.source, m.DataSource):
            return self.source
        elif isinstance(self.source, m.ArtifactPaths):
            return m.DataSource(artifact_paths=self.source)
        return m.DataSource(artifact_paths=self.source._build_artifact_paths())

    def _build_data(self) -> m.Data:
        """Builds the generated `Data` template."""
        return m.Data(
            source=self._build_source(),
            transformation=list(map(lambda expr: m.TransformationStep(expression=str(expr)), self.transformations)),
        )

    def _build_template(self) -> m.Template:
        """Builds the generated `Template` from the fields of `Data`."""
        return m.Template(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            archive_location=self.archive_location,
            automount_service_account_token=self.automount_service_account_token,
            data=self._build_data(),
            executor=self.executor,
            fail_fast=self.fail_fast,
            host_aliases=self.host_aliases,
            init_containers=self._build_init_containers(),
            inputs=self._build_inputs(),
            outputs=self._build_outputs(),
            memoize=self.memoize,
            metadata=self._build_metadata(),
            name=self.name,
            node_selector=self.node_selector,
            plugin=self.plugin,
            priority=self.priority,
            priority_class_name=self.priority_class_name,
            retry_strategy=self.retry_strategy,
            scheduler_name=self.scheduler_name,
            security_context=self.pod_security_context,
            service_account_name=self.service_account_name,
            sidecars=self._build_sidecars(),
            synchronization=self.synchronization,
            timeout=self.timeout,
            tolerations=self.tolerations,
        )


__all__ = ["Data"]



"""The auth module of Hera consists of authentication related implementation.

The module provides basic functionality such as token generation via the Argo CLI, the basis of token generation for
implementing a client token generator, etc.
"""

import os
import random
import shutil
import subprocess


class TokenGenerator:
    """A token generator can be used to generate tokens for authentication with Argo Workflows/Events APIs.

    A token generator can be set for invocation on the Hera global config via
    `hera.shared.global_config.token`.
    """

    def __call__(self) -> str:
        """Generates an authentication token for use with Argo Workflows/Events APIs."""
        raise NotImplementedError("Implement me")


class ArgoCLITokenGenerator(TokenGenerator):
    """A token generator that uses the Argo CLI to generate a token.

    Notes:
    -----
    This involves invoking the Argo CLI, which means that the Argo CLI must be installed on the machine.
    An exception is raised if this is not the case.

    Raises:
    ------
    RuntimeError
        If the Argo CLI is not installed.
    """

    def __call__(self) -> str:
        """Executes the call necessary to generate the token and returns the token as a string."""
        if shutil.which("argo") is None:
            raise RuntimeError(
                "The Argo CLI is not installed. "
                "See `https://argoproj.github.io/argo-workflows/walk-through/argo-cli/` for more information"
            )

        token = subprocess.check_output("argo auth token".split()).strip().decode()
        if token.startswith("Bearer "):
            token = token[7:]
        return token



"""Hera classes."""
# [DO NOT EDIT MANUALLY]
# Auto-generated by Hera via `make init-files`
# In order to add objects to the hera.workflows namespace
# add them to the __all__ list in the relevant module.
# Hera submodules should not use `from hera.workflows import X`
# themselves, as it introduces a circular dependency.

from hera.workflows.archive import ArchiveStrategy, NoneArchiveStrategy, TarArchiveStrategy, ZipArchiveStrategy
from hera.workflows.artifact import (
    Artifact,
    ArtifactLoader,
    ArtifactoryArtifact,
    AzureArtifact,
    GCSArtifact,
    GitArtifact,
    HDFSArtifact,
    HTTPArtifact,
    OSSArtifact,
    RawArtifact,
    S3Artifact,
)
from hera.workflows.cluster_workflow_template import ClusterWorkflowTemplate
from hera.workflows.container import Container
from hera.workflows.container_set import ContainerNode, ContainerSet
from hera.workflows.cron_workflow import CronWorkflow
from hera.workflows.dag import DAG
from hera.workflows.data import Data
from hera.workflows.env import ConfigMapEnv, Env, FieldEnv, ResourceEnv, SecretEnv
from hera.workflows.env_from import ConfigMapEnvFrom, SecretEnvFrom
from hera.workflows.exceptions import (
    InvalidDispatchType,
    InvalidTemplateCall,
    InvalidType,
    NodeNameConflict,
    TemplateNameConflict,
)
from hera.workflows.http_template import HTTP
from hera.workflows.io import Input, Output
from hera.workflows.metrics import Counter, Gauge, Histogram, Label, Metric, Metrics
from hera.workflows.operator import Operator
from hera.workflows.parameter import Parameter
from hera.workflows.resource import Resource
from hera.workflows.resources import Resources
from hera.workflows.retry_strategy import RetryPolicy, RetryStrategy
from hera.workflows.script import InlineScriptConstructor, RunnerScriptConstructor, Script, ScriptConstructor, script
from hera.workflows.service import WorkflowsService
from hera.workflows.steps import Parallel, Step, Steps, parallel
from hera.workflows.suspend import Suspend
from hera.workflows.task import Task, TaskResult
from hera.workflows.template_set import TemplateSet
from hera.workflows.user_container import UserContainer
from hera.workflows.volume import (
    AccessMode,
    AWSElasticBlockStoreVolumeVolume,
    AzureDiskVolumeVolume,
    AzureFileVolumeVolume,
    CephFSVolumeVolume,
    CinderVolume,
    ConfigMapVolume,
    CSIVolume,
    DownwardAPIVolume,
    EmptyDirVolume,
    EphemeralVolume,
    ExistingVolume,
    FCVolume,
    FlexVolume,
    FlockerVolume,
    GCEPersistentDiskVolume,
    GitRepoVolume,
    GlusterfsVolume,
    HostPathVolume,
    ISCSIVolume,
    NFSVolume,
    PhotonPersistentDiskVolume,
    PortworxVolume,
    ProjectedVolume,
    QuobyteVolume,
    RBDVolume,
    ScaleIOVolume,
    SecretVolume,
    StorageOSVolume,
    Volume,
    VsphereVirtualDiskVolume,
)
from hera.workflows.workflow import Workflow
from hera.workflows.workflow_status import WorkflowStatus
from hera.workflows.workflow_template import WorkflowTemplate

__all__ = [
    "AWSElasticBlockStoreVolumeVolume",
    "AccessMode",
    "ArchiveStrategy",
    "Artifact",
    "ArtifactLoader",
    "ArtifactoryArtifact",
    "AzureArtifact",
    "AzureDiskVolumeVolume",
    "AzureFileVolumeVolume",
    "CSIVolume",
    "CephFSVolumeVolume",
    "CinderVolume",
    "ClusterWorkflowTemplate",
    "ConfigMapEnv",
    "ConfigMapEnvFrom",
    "ConfigMapVolume",
    "Container",
    "ContainerNode",
    "ContainerSet",
    "Counter",
    "CronWorkflow",
    "DAG",
    "Data",
    "DownwardAPIVolume",
    "EmptyDirVolume",
    "Env",
    "EphemeralVolume",
    "ExistingVolume",
    "FCVolume",
    "FieldEnv",
    "FlexVolume",
    "FlockerVolume",
    "GCEPersistentDiskVolume",
    "GCSArtifact",
    "Gauge",
    "GitArtifact",
    "GitRepoVolume",
    "GlusterfsVolume",
    "HDFSArtifact",
    "HTTP",
    "HTTPArtifact",
    "Histogram",
    "HostPathVolume",
    "ISCSIVolume",
    "InlineScriptConstructor",
    "Input",
    "InvalidDispatchType",
    "InvalidTemplateCall",
    "InvalidType",
    "Label",
    "Metric",
    "Metrics",
    "NFSVolume",
    "NodeNameConflict",
    "NoneArchiveStrategy",
    "OSSArtifact",
    "Operator",
    "Output",
    "Parallel",
    "Parameter",
    "PhotonPersistentDiskVolume",
    "PortworxVolume",
    "ProjectedVolume",
    "QuobyteVolume",
    "RBDVolume",
    "RawArtifact",
    "Resource",
    "ResourceEnv",
    "Resources",
    "RetryPolicy",
    "RetryStrategy",
    "RunnerScriptConstructor",
    "S3Artifact",
    "ScaleIOVolume",
    "Script",
    "ScriptConstructor",
    "SecretEnv",
    "SecretEnvFrom",
    "SecretVolume",
    "Step",
    "Steps",
    "StorageOSVolume",
    "Suspend",
    "TarArchiveStrategy",
    "Task",
    "TaskResult",
    "TemplateNameConflict",
    "TemplateSet",
    "UserContainer",
    "Volume",
    "VsphereVirtualDiskVolume",
    "Workflow",
    "WorkflowStatus",
    "WorkflowTemplate",
    "WorkflowsService",
    "ZipArchiveStrategy",
    "parallel",
    "script",
]
